#!/usr/bin/env python
#----------------------------------------------------------------------------
# Name:         OritatamiSim.py
# Purpose:      A basic program for drawing shapes, scaling and rotating them,
#               and concatenating sets of bead sequences to create an Oritatami
#               system which self-assembles the shape (at constant scale factor
#               and with a constant set of bead types, but with a sequence which
#               grows linearly with the size of the shape).
#
# Author:       Matt Patitz
#
# Created:      
# RCS-ID:       
# Copyright:    
# Licence:      
#----------------------------------------------------------------------------

"""
This program can load the definition of a shape system from a file and
generate an Oritatami system which self-assembles a scaled version of it.

Requirements:
    This application requires Python 2.x due to its dependency on wxPython

Dependencies:
    wxPython    (download installer for your platform here:  https://www.wxpython.org/download.php)
    pil         (via pillow, install using 'pip install pillow', pip can be found in your Python27\Scripts dir)
"""


"""
MJP TODO list:
    - verify that each node only has two edges after creating Hamiltonian cycle
    - improve fast forward functionality to be multi-threaded
    - BUG:  Toggling the grid and zooming can cause the grid to be shifted partially off of the visible window area
"""


import wx
import wx.richtext as rt
import os
import random
import math
import re
import sys
import thread
from collections import deque
import itertools
import wx.aui
import time
import threading
import wx.lib.scrolledpanel as scrolled
from wx.lib.agw.shapedbutton import SButton, SBitmapButton
import wx.lib.agw.cubecolourdialog as CCD
import wx.html
from LookupTable import *
#import ast # TODO: This was needed for debugging remove when done using 

DEFINITION_FILE_OASS = "input/gadgetDefinitions.oass"
DEFINITION_FILE_ORIT = "input/gadgetDefinitions.orit"
DEFINITION_FILE_TXT  = "input/gadgetDefinitions.txt"

RAND_SEED = -1

#DocumentationFileList = [ "Overview", "System Parameters", "File Formats", "Tips and Tricks"]


GRIDSTARTSIZE = 30
GRIDGROWTHINC = 30
TRISIZE = 60

# IDs
ORIT_ID_FR             = 100
ORIT_ID_REV            = 101
ORIT_ID_STOP           = 102
ORIT_ID_FOR            = 103
ORIT_ID_FF             = 104
ORIT_ID_NEW            = 105
ORIT_ID_LOAD           = 106
ORIT_ID_SAVE           = 107
ORIT_ID_ZOOMIN         = 108
ORIT_ID_ZOOMOUT        = 109
ORIT_ID_MODE_PAN_ZOOM  = 110
ORIT_ID_MODE_DRAW_BEAD = 111
ORIT_ID_MODE_DRAW_LINK = 112
ORIT_ID_MODE_DRAW_BOND = 113
ORIT_ID_RESET          = 114
ORIT_ID_RESET_BEADSEQ  = 115
ORIT_ID_RELOAD_BEADSEQ = 116
ORIT_ID_CANCEL_BEADSEQ = 117
ORIT_ID_OK_NOCHECK     = 118
ORIT_ID_OK_CHECK       = 119
ORIT_ID_CANCEL         = 120
ORIT_ID_COMPILE        = 121
ORIT_ID_GENOS          = 122

ORIT_MODE_PAN_ZOOM     = 0
ORIT_MODE_DRAW_BEAD    = 1
ORIT_MODE_DRAW_LINK    = 2
ORIT_MODE_DRAW_BOND    = 3

DEBUG_MSG = 0
MSG_MSG   = 1
IMP_MSG   = 2
WARN_MSG  = 3
ERROR_MSG = 4
DEBUG_MSG_LEVEL = WARN_MSG

# Result statuses for stepping forward in simulation
ORI_STAT_OK         = 0
ORI_STAT_NO_SPACE   = 1
ORI_STAT_NO_BINDING = 2
ORI_STAT_NONDET     = 3
ORI_STAT_MAX_FLOPPY = 4
ORI_STAT_ERROR      = 5


def PrintDebug(msg, level):
    if DEBUG_MSG_LEVEL <= level:
        print msg

colours = [
    "BLACK",
    "BLUE",
    "BLUE VIOLET",
    "BROWN",
    "CYAN",
    "DARK GREY",
    "DARK GREEN",
    "GOLD",
    "GREY",
    "GREEN",
    "MAGENTA",
    "NAVY",
    "PINK",
    "RED",
    "SKY BLUE",
    "VIOLET",
    "YELLOW",
    ]

N = 0
E = 1
S = 2
W = 3
ALLDIRS = 4
DIRSTRS = ("N", "E", "S", "W")

SHOW_GRID_AND_COORDS = 0
SHOW_GRID_ONLY       = 1
SHOW_NO_GRID         = 2

assertMode = wx.PYAPP_ASSERT_DIALOG
##assertMode = wx.PYAPP_ASSERT_EXCEPTION

globalApp = None

class ONodeType:
    def __init__(self, name, color, bondToTypes):
        self.name = name
        self.color = color
        
    def __repr__(self):
        return self.name
    
    def __str__(self):
        return self.name

    def Output(self):
        outStr = "['" + self.name + "', '" + str(self.color) + "]"
        return outStr
    
ShapeNodeType = ONodeType("shapeNode", "BLUE", "")
ScaledNodeType = ONodeType("scaledNode", "BLACK", "")
GadgetNodeType = ONodeType("gadgetNode", "YELLOW", "")

class ONode:

    def __init__(self, otype, coords=None, parent=None):
        self.type = otype
        self.coords = coords
        self._bonds = []
        self.parent = parent
        self.child = None

    def __str__(self):
        outStr = "ONode[coords " + str(self.coords) + ", type " + str(self.type) + "]"
        return outStr

    def __repr__(self):
        outStr = "ONode[" + str(self.coords) + "]"
        return outStr


ONBRS = [(0,0),(0,1),(1,0),(1,-1),(0,-1),(-1,0),(-1,1)]
ONBRINDEX = {(0,0): 0,
             (0,1): 1,
             (1,0): 2,
             (1,-1): 3,
             (0,-1): 4,
             (-1,0): 5,
             (-1,1): 6}

NBRFLIP = {(1,2):   (4,5),
           (2,3):   (5,6),
           (3,4):   (6,1),
           (4,5):   (1,2),
           (5,6):   (2,3),
           (6,1):   (3,4)}

SCALEDNBRS = [(0,0),(1,1),(2,-1),(1,-2),(-1,-1),(-2,1),(-1,2)]

SHAREDEDGES = {(1,1):   [(0,1), (1,0)],
               (2,-1):  [(1,0), (1,-1)],
               (1,-2):  [(1,-1), (0,-1)],
               (-1,-1): [(0,-1), (-1,0)],
               (-2,1):  [(-1,0), (-1,1)],
               (-1,2):  [(-1,1), (0,1)]}

NbrCount1 = {str([(1,2)]): {"Straight" : {"Add" : [(2,3),(3,4),(4,5),(5,6),(6,0),(0,1)],
                                   "NbrAdd" : [], 
                                   "NbrRemove" : [(1,2)]},
                            "V" : {"Add" : [(2,3),(3,4),(4,5),(5,6),(6,0),(0,1)],
                                   "NbrAdd" : [],
                                   "NbrRemove" : [(1,0),(0,2)]}
                            }
            }

NbrCount2 = {str([(1,2),(2,3)]): "foo",
             str([(1,2),(3,4)]): "foo2",
             str([(1,2),(4,5)]): "foo3"}

EdgePatterns = {0: [(0,6),(6,1),(1,2),(2,3),(3,4),(4,5),(5,0)],
                1: NbrCount1,
                2: NbrCount2,
                3: None,
                4: None,
                5: None,
                6: None}

def RotateEdges(edgeList, rotAmount):
    rotatedEdges = []
    for edge in edgeList:
        v1 = edge[0]
        if v1 > 0 and v1 < 7:
            v1 = ((v1 + (rotAmount-1)) % 6) + 1
        v2 = edge[1]
        if v2 > 0 and v2 < 7:
            v2 = ((v2 + (rotAmount-1)) % 6) + 1
        rotatedEdge = (v1, v2)
        rotatedEdges.append(rotatedEdge)
    return rotatedEdges

def RotateNbrGadgetNum(nbrNum, rot):
    if rot < 0:
        rot = rot + 6
    if nbrNum == 0:
        return 0
    else:
        nbrNum = nbrNum + rot
        if nbrNum > 6:
            nbrNum = (nbrNum % 6)
            if nbrNum == 0:
                nbrNum = 6
        return nbrNum

class NodeGadget:
    def __init__(self, coords):
        self.coords = coords
        self.edges = []
        
    def __eq__(self, other):
        return isinstance(other, NodeGadget) and self.coords == other.coords

    def __hash__(self):
        return hash(self.coords)
    
    def AddEdges(self, edgeList):
        for e in edgeList:
            offset1 = ONBRS[e[0]]
            offset2 = ONBRS[e[1]]
            edge = [offset1, offset2]
            self.edges.append(edge)
            
    def RemoveEdges(self, edgeList):
        for e in edgeList:
            offset1 = ONBRS[e[0]]
            offset2 = ONBRS[e[1]]
            edge = [offset1, offset2]
            if edge in self.edges:
                self.edges.remove(edge)
                #print("Removed edge " + str(edge))
            else:
                edge = [offset2, offset1]
                if edge in self.edges:
                    self.edges.remove(edge)
                else:
                    globalApp.SetOutputMessage("ERROR: Trying to remove nonexistent edge!!")
            
    def GetNbrBorderPattern(self, nbrGadget):
        #print("GetNbrBorderPattern for gadget at " + str(self.coords) + " and nbr at " + str(nbrGadget.coords))
        nbr = nbrGadget.coords

        # Get the boundary edges from the neighbor's perspective
        nbrOffset = (self.coords[0]-nbr[0], self.coords[1]-nbr[1])
        boundaryEdge = SHAREDEDGES[nbrOffset]
        sharedVerts = (ONBRINDEX[boundaryEdge[0]], ONBRINDEX[boundaryEdge[1]])
        fullEdge = [ONBRS[sharedVerts[0]],ONBRS[sharedVerts[1]]]
        fullEdgeReversed = [fullEdge[1], fullEdge[0]]
        V = [[ONBRS[sharedVerts[0]],(0,0)],[(0,0),ONBRS[sharedVerts[1]]]]
        VReversed = [[(0,0),ONBRS[sharedVerts[0]]],[ONBRS[sharedVerts[1]],(0,0)]]
        #print("Nbr's shared verts: " + str(sharedVerts) + ", full edge: " + str(fullEdge) + ", and nbr's edges " + str(nbrGadget.edges))
        #print("V: " + str(V))
        flippedVerts = NBRFLIP[(sharedVerts[0], sharedVerts[1])]
        if (fullEdge in nbrGadget.edges) or (fullEdgeReversed in nbrGadget.edges):
            # Return just this edge, from the caller's perspective
            #print("HAS EDGE")
            return [flippedVerts]
        elif (V[0] in nbrGadget.edges and V[1] in nbrGadget.edges) or (VReversed[0] in nbrGadget.edges and VReversed[1] in nbrGadget.edges):
            #print("HAS V")
            return [(flippedVerts[0], 8), (8, flippedVerts[1])]
        else:
            return [(9,9)]
            #globalApp.SetOutputMessage("Encountered unrecognized edge pattern on gadget border.  Aborting.")
            #print("HAS NEITHER: Gotta figure out what to return...")
            #return None

    def DrawPolygon(self, pdc, canvas):
        pdc.SetBrush(canvas.ColoredBrush("YELLOW"))
        pdc.SetPen(canvas.gridPen)
        polygon = []
        for n in ONBRS[1:]:
            no = (self.coords[0]+n[0], self.coords[1]+n[1])
            px = canvas.GetPixelsForCoords(no[0], no[1])
            polygon.append(px)
        polygon.append(polygon[0])
        pdc.DrawPolygon(polygon)

    def Draw(self, pdc, canvas):
        #print("time to draw gadget with edges " + str(self.edges))
        pdc.SetBrush(canvas.ColoredBrush("YELLOW"))
        pdc.SetPen(canvas.cyclePen)
        for e in self.edges:
            loc1 = (self.coords[0]+e[0][0], self.coords[1]+e[0][1])
            loc2 = (self.coords[0]+e[1][0], self.coords[1]+e[1][1])
            px1 = canvas.GetPixelsForCoords(loc1[0], loc1[1])
            px2 = canvas.GetPixelsForCoords(loc2[0], loc2[1])
            pdc.DrawLine(px1[0], px1[1], px2[0], px2[1])
            pdc.DrawCirclePoint((px1[0], px1[1]), 10)
            pdc.DrawCirclePoint((px2[0], px2[1]), 10)
        pdc.SetBrush(canvas.ColoredBrush("BLACK"))
        px1 = canvas.GetPixelsForCoords(self.coords[0], self.coords[1])
        pdc.DrawCirclePoint((px1[0], px1[1]), 10)
        return True

class GadgetSet:
    def __init__(self):
        self.gadgets = dict()
    
    def Add(self, nodeGadget):
        self.gadgets[nodeGadget.coords] = nodeGadget

    def HasLoc(self, coords):
        if coords in self.gadgets.keys():
            return True
        return False

    def Get(self, coords):
        if self.HasLoc(coords):
            return self.gadgets[coords]
        return None

    def GetNbrs(self, coords):
        if self.HasLoc(coords):
            snbrList = []
            for sn in SCALEDNBRS[1:]:
                snbr = (coords[0]+sn[0], coords[1]+sn[1])
                if self.HasLoc(snbr):
                    snbrList.append(self.Get(snbr))
            return snbrList
        return None
    
    def GetAbstractNbrPattern(self, coords):
        if self.HasLoc(coords):
            nbrPattern = []
            for sn in SCALEDNBRS[1:]:
                abstractPattern = "_"
                snbr = (coords[0]+sn[0], coords[1]+sn[1])
                if self.HasLoc(snbr):
                    pattern = self.Get(coords).GetNbrBorderPattern(self.Get(snbr))
                    #print("got pattern: " + str(pattern))
                    if (9,9) in pattern:
                        abstractPattern = "D"
                    else:
                        if len(pattern) == 1:
                            abstractPattern = "S"
                        else:
                            abstractPattern = "V"
                nbrPattern.append(abstractPattern)
            return nbrPattern
        return None
    
    def ExtendCycle(self, coords, rot, reflect, changeList):
        #print("In ExtendCycle about to implement changes: " + str(changeList))
        for gadgetChanges in changeList:
            nbrNum = RotateNbrGadgetNum(gadgetChanges[0], rot)
            addList = gadgetChanges[1]
            removeList = gadgetChanges[2]
            
            nbrOffset = SCALEDNBRS[nbrNum]
            gadgetCoords = (coords[0]+nbrOffset[0], coords[1]+nbrOffset[1])

            if not self.HasLoc(gadgetCoords):
                globalApp.SetOutputMessage("ERROR: In ExtendCycle, no gadget at " + str(gadgetCoords))
                return False

            #print("  in loc " + str(gadgetCoords) + ":")
                 
            rotAdd = RotateEdges(addList, rot)
             
            #print("        adding: " + str(addList) + " rotated: " +  str(rotAdd))
            self.Get(gadgetCoords).AddEdges(rotAdd)
            
            rotRem = RotateEdges(removeList, rot)

            #print("       removing: " + str(removeList) + " rotated: " + str(rotRem))
            self.Get(gadgetCoords).RemoveEdges(rotRem)
            
        return True

        
class ONodeGrid:
    def __init__(self, width, height):
        self.MakeGrid(width, height)
        self.delayLength = 2    # Default delay of 2
        self.arity = 1          # Maximum number of bonds for a single bead
        self.seed = []
        self.beadCycle = []
        self.origBeadCycle = []
        self.currCyclePos = 0
        self.lastBeadLoc = None
        self.lastSeedLoc = None
        self.floppySegment = deque([])
        self.ONodeTypes = dict()
        self.shapePoints = []
        self.orderedList = []
        self.scaledAndRotatedList = []
        self._numBeads = 0
        
        # Hard code the bead types
        self.ONodeTypes["shapeNode"] = ShapeNodeType
        self.ONodeTypes["scaledNode"] = ScaledNodeType
        self.ONodeTypes["gadgetNode"] = GadgetNodeType

    def GetNumBeads(self):
        return self._numBeads

    def GetWidth(self):
        return self._width
    
    def GetHeight(self):
        return self._height

    def MakeGrid(self, width, height, bResetBeadNum=False):
        self._width = width
        self._height = height
        self._grid = [[None for x in range(width)] for y in range(height)]
        self.minX = 0 - int(math.floor(width/2))
        self.maxX = int(math.floor(width/2))
        self.minY = 0 - int(math.floor(height/2))
        self.maxY = int(math.floor(height/2))
        if (width % 2) == 0:
            self.minX = self.minX + 1
        if (height % 2) == 0:
            self.minY = self.minY + 1
        if bResetBeadNum == True:
            self._numBeads = 0

    def GetAllGadgets(self):
        retList = []
        for i in range(self._width):
            for j in range(self._height):
                loc = self._grid[i][j]
                if not loc is None:
                    coords = (i+self.minX, j+self.minY)
                    retList.append(coords)
        return retList

    def GetLoc(self, coords):
        if coords[0] < self.minX or coords[0] > self.maxX or coords[1] < self.minY or coords[1] > self.maxY:
            print("Growing grid because of access to " + str(coords))
            self.GrowGrid()
        return self._grid[coords[0]-self.minX][coords[1]-self.minY]
    
    def SetLoc(self, coords, onode):
        if coords[0] < self.minX or coords[0] > self.maxX or coords[1] < self.minY or coords[1] > self.maxY:
            self.GrowGrid()
        self._grid[coords[0]-self.minX][coords[1]-self.minY] = onode
        return True

    def GrowGrid(self):
        oldgrid = self._grid
        oldwidth = self._width
        oldheight = self._height
        oldminx = self.minX
        oldminy = self.minY
        self.MakeGrid(oldwidth + GRIDGROWTHINC, oldheight + GRIDGROWTHINC)
        for i in range(oldwidth):
            for j in range(oldheight):
                oldnode = oldgrid[i][j]
                self.SetLoc((i + oldminx, j + oldminy), oldnode)

    def AddNode(self, otype, coords, parentCoords, bWarnOnNoParent=True):
        childNode = ONode(otype, coords)
        if not self.GetLoc(coords) is None:
            print("ERROR: Trying to add another node at location " + str(coords))
            return
        self.SetLoc(coords, childNode)
        self.shapePoints.append(coords)
        self._numBeads = self._numBeads + 1
        
    def RemoveNode(self, coords):
        node = self.GetLoc(coords)
        if not node is None:
            
            self.shapePoints.remove(coords)
            self.SetLoc(coords, None)
            self._numBeads = self._numBeads - 1
        else:
            print("ERROR: Trying to remove nonexistent node at location" + str(coords))

    def AreNbrLocs(self, coords1, coords2):
        if (coords1 is None) or (coords2 is None):
            return False
        for n in ONBRS[1:]:
            nbrCoords = (coords1[0]+n[0], coords1[1]+n[1])
            if nbrCoords == coords2:
                return True
        return False

    def GetNbrLocs(self, coords, bEmpty=False):
        retlist = []
        for n in ONBRS[1:]:
            nbrCoords = (coords[0]+n[0], coords[1]+n[1])
            if bEmpty == False:
                retlist.append(nbrCoords)
            elif self.GetLoc(nbrCoords) is None:
                retlist.append(nbrCoords)
        return retlist

    def GetOccupiedNbrLocs(self, coords, bSkipParent=False, maxBonds=0):
        retlist = []
        for n in ONBRS[1:]:
            nbrCoords = (coords[0]+n[0], coords[1]+n[1])
            nbrNode = self.GetLoc(nbrCoords)
            if not nbrNode is None:
                #if (bSkipParent == False) or (nbrNode.child is None) or (not nbrNode.child.coords == coords):
                #    if nbrNode.NumBonds() <= maxBonds:
                retlist.append(nbrNode)
        return retlist

    def GetTopLeft(self):
        # Get leftmost point of the top row
        for i in reversed(range(self._height)):
            for j in range(self._width):
                n = self._grid[j][i]
                if not n is None:
                    y = i + self.minY
                    x = j + self.minX
                    return (x,y)
        return None

    def GetOrderedList(self):
        topLeft = self.GetTopLeft()     
        if topLeft is None:
            print("No points in the shape!")
            return None
        print("Topleft: " + str(topLeft))
        
        orderedPoints = [topLeft]
        n = 0
        while n < len(orderedPoints):
            nbrs = self.GetOccupiedNbrLocs(orderedPoints[n])
            for nbr in nbrs:
                if not nbr.coords in orderedPoints:
                    orderedPoints = orderedPoints + [nbr.coords]
            n = n + 1
        return orderedPoints

    def LoadOritatamiSystem(self, origFileName, origFileNameFullPath=None):
        fileName = origFileName
        if not origFileNameFullPath is None:
            fileName = origFileNameFullPath
        if len(fileName) < 1:
            return False
        elif os.path.isabs(fileName) == False:
            fileName = "./" + fileName
            fileName = os.path.abspath(fileName)

        try:
            os.stat(fileName)
            rFile = open(fileName, "r")
            print("Using oritatami system definition file '" + fileName + "'")
            lines = rFile.readlines()
            rFile.close()
    
            self.ONodeTypes = dict()
    
            for line in lines:
                if line.startswith("BeadTypes="):
                    try:
                        beadTypes = eval(line[len("BeadTypes="):])
                        for bead in beadTypes:
                            beadType = ONodeType(bead[0], bead[1], bead[2])
                            self.ONodeTypes[bead[0]] = beadType
                        self.FixBeadBondSymmetry(True)
                    except:
                        globalApp.ShowErrorDialog("Invalid bead type definitions.  See messages window.")
                        globalApp.SetOutputMessage("Invalid line '" + line + "'.  Bead types should be a comma separated list surrounded by '[' and ']'. Each bead type definition should similar to the following:  ['Name', 'Color', ['Name2', 'Name3']] where Name2 and Name3 are in a list of bead types that this type binds to.")
                        return False
                elif line.startswith("DelayLength="):
                    dlLine = line[len("DelayLength="):].strip()
                    self.delayLength = int(dlLine)
                    print("DelayLength: " + str(self.delayLength))
                    #if (self.delayLength > 2):
                        #MJP TODO: fix this issue!
                        #print("\n\nWARNING:  For delay longer than 2, newly extended beads could possibly bind to each other, but this is not yet implemented!!!\n")
                elif line.startswith("Arity="):
                    arityLine = line[len("Arity="):].strip()
                    self.arity = int(arityLine)
                    print("Arity: " + str(self.arity))
                elif line.startswith("BeadCycle="):
                    bcLine = line[len("BeadCycle="):].strip()
                    self.beadCycle = []
                    self.origBeadCycle = []
                    try:
                        self.beadCycle = eval(bcLine)
                        self.origBeadCycle = self.beadCycle
                    except:
                        globalApp.ShowErrorDialog("Invalid bead cycle definition.  See messages window.")
                        globalApp.SetOutputMessage("Invalid line '" + line + "'.  Bead cycle should be a comma separated list of bead type names surrounded by '[' and ']', similar to the following:  ['Name1', 'Name2', 'Name3'].")
                        return False
                    print("beadCycle: " + str(self.beadCycle))

            self.InitializeFloppySegment()
            globalApp.ResetBreakpoints()
            globalApp.systemDefinitionFile = origFileName
            globalApp.systemDefinitionFileFullPath = origFileNameFullPath
            globalApp.SetOutputMessage("Loaded bead type definitions from file '" + fileName + "'")
            print("Loaded bead set: " + str(self.ONodeTypes))
            return True;
        except:
            globalApp.ShowErrorDialog("\nERROR: Invalid oritatami system definition file: '" + fileName + "'\n")
            return False

    def LoadOritatamiAssembly(self, origFileName, bForceDefLoad=False):
        fileName = origFileName
        if len(fileName) < 1:
            return False
        elif os.path.isabs(fileName) == False:
            fileName = "./" + fileName
            fileName = os.path.abspath(fileName)

        lastLoc = None
        try:
            os.stat(fileName)
            rFile = open(fileName, "r")
            print("Using shape definition file '" + fileName + "'")
            lines = rFile.readlines()
            rFile.close()
            
            for line in lines:
                if line.startswith("Assembly="):
                    if self.LoadOritatamiAssemblyHelper(line) == False:
                        globalApp.SetOutputMessage("ERROR: Invalid shape definition file: '" + fileName + "'")
                        return False
        except:
            globalApp.ShowErrorDialog("\nERROR: Invalid shape definition file: '" + fileName + "'")
            return False
                
        globalApp.seedDefinitionFile = origFileName
        globalApp.SetOutputMessage("Loaded shape definition from file '" + globalApp.seedDefinitionFile + "'")
        return True

    def LoadOritatamiAssemblyHelper(self, seedString):
        seedString = seedString[len("Assembly="):].strip()
        try:
            self.shapePoints = []
            seedArr = eval(seedString)
            for b in seedArr:
                if b[0] in self.ONodeTypes:
                    self.shapePoints.append(b[1])
                else:
                    globalApp.SetOutputMessage("ERROR: Trying to add a bead of undefined type '" + str(b[0]) + "' at " + str(b[1]))
        except:
            globalApp.ShowErrorDialog("\nERROR: Invalid shape definition file.")
            return False
        return self.ResetToShapeDef()
        
    def ResetToShapeDef(self):
        print("Shape definition: " + str(self.shapePoints))
        self.MakeGrid(self._width, self._height, True)
        savedShapePoints = copy.deepcopy(self.shapePoints)
        self.shapePoints = []
        for p in savedShapePoints:
            self.AddNode(ShapeNodeType.name, p, None, False)

        return True

#----------------------------------------------------------------------------

class MyCanvas(wx.ScrolledWindow):
    def __init__(self, parent, idIn, frame, size, app):
        wx.ScrolledWindow.__init__(self, parent, idIn, (0, 0), size=size, style=wx.SUNKEN_BORDER)

        self.frame = frame
        self.app = app
        self.x = self.y = 0
        self.scale = 1.0
        self.origin = wx.Point()
        self.DrawQueue = []
        self.RemoveQueue = []
        self.objids = []
        self.dictIdToType = {}
        self.boundsdict = {}
        self.keepGoing = False
        self.showMouseOver = True
        self.breakpoints = []
        self.currentBreakpoint = -1
        self.breakOnNondeterminism = False
        self.confirmOverArity = True
        self.confirmInvalidBonds = True
        self._lock = threading.Lock()
        
        self.showGrid = SHOW_GRID_AND_COORDS
        
        self.mouseMode = ORIT_MODE_PAN_ZOOM
        self.drawingModeBeadType = ShapeNodeType

        # create a PseudoDC to record our drawing
        self.pdc = wx.PseudoDC()
        self.pen_cache = {}
        self.brush_cache = {}
        self.gridPen = self.ColoredPen("RED", 1)
        self.axesPen = self.ColoredPen("RED", 1)
        self.bondPen = self.ColoredPen("RED", 2)
        self.linkPen = self.ColoredPen("BLACK", 4)
        self.cyclePen = self.ColoredPen("RED", 8)
        self.grabbingCursorImg = wx.Image("images/grabbing.gif", wx.BITMAP_TYPE_GIF)

        self.SetBackgroundColour("WHITE")
        
        self.sideLen = TRISIZE
        self.triangleHeight = math.tan(math.radians(60)) * 0.5 * self.sideLen

        self.gridWidth = GRIDSTARTSIZE
        self.gridHeight = GRIDSTARTSIZE

        self.SetScrollRate(1,1)

        self.Bind(wx.EVT_PAINT, self.OnPaint)
        self.Bind(wx.EVT_ERASE_BACKGROUND, lambda x:None)
        self.Bind(wx.EVT_MOUSE_EVENTS, self.OnMouse)
        self.Bind(wx.EVT_CHAR,self.OnChar)
        
        # vars for handling mouse clicks
        self.dragid = -1
        self.lastpos = (0,0)
        self.originalpos = (0,0)
        
        self.dragStart   = None
        self.linkStart   = None
        self.linkEnd     = None
        self.posSelected = None
        self.negSelected = None
        self.lastCurrent = None
        
        self.SetRandomSeed(RAND_SEED)

        self.onodegrid = ONodeGrid(GRIDSTARTSIZE, GRIDSTARTSIZE)
        self.InitializeDrawing(self.pdc)
        #self.ScrollToCoords((0,0))

    def DoLoad(self, bAssembly=True, bSystem=False):
        typeString = "Shape definition files (*.opts)|*.opts"
        openFileDialog = wx.FileDialog(self, "Open", "", "", 
                                       typeString,
                                       wx.FD_OPEN | wx.FD_FILE_MUST_EXIST)
        openFileDialog.ShowModal()
        fileName = openFileDialog.GetPath()
        openFileDialog.Destroy()
        if (len(fileName) > 0):

            self.onodegrid = ONodeGrid(GRIDSTARTSIZE,GRIDSTARTSIZE)
            self.onodegrid.LoadOritatamiAssembly(fileName, True)

                #if len(self.onodegrid.ONodeTypes) > 0:
                    # Randomly set the current drawing bead type to one of the bead types
                    #self.drawingModeBeadType = self.onodegrid.ONodeTypes[self.onodegrid.ONodeTypes.keys()[0]]
                #else:
            self.InitializeDrawing(self.pdc)
            self.Refresh()
            self.ScrollToCoords((0,0))

    def DoSaveShapeDef(self, bSaveAs=False):
        fileName = globalApp.systemDefinitionFileFullPath
        if fileName is None or bSaveAs == True:
            saveFileDialog = wx.FileDialog(self, "Save Shape Definition As", "", "", 
                                           "Shape definition files (*.opts)|*.opts",
                                           wx.FD_SAVE | wx.FD_OVERWRITE_PROMPT)
            if not saveFileDialog.ShowModal() == wx.ID_OK:
                globalApp.ShowErrorDialog("You must provide a file name to save the shape definition to.")
                return False
            fileName = saveFileDialog.GetPath()
            globalApp.systemDefinitionFile = fileName
            saveFileDialog.Destroy()

        globalApp.systemDefinitionFileFullPath = fileName
        if os.path.isabs(fileName) == False:
            fileName = "./" + fileName
            fileName = os.path.abspath(fileName)

        try:
            btStr = "["
            for bt in self.onodegrid.ONodeTypes:
                if len(btStr) > 1:
                    btStr = btStr + ", "
                btStr = btStr + self.onodegrid.ONodeTypes[bt].Output()
            btStr = btStr + "]"
            btLine = "BeadTypes=" + btStr
            #print("btLine: " + btLine)
            assemblyString = "Assembly="
            #lastBeadString = "LastBead=" + str(self.onodegrid.lastBeadLoc)
            assemblyList = self.GetAssemblyList(bOnlySeed=False)
            if len(assemblyList) == 0:
                globalApp.ShowErrorDialog("The shape definition is empty.  It must contain at least one point.")
                return False
            assemblyString = assemblyString + str(assemblyList)

            wFile = open(fileName, "w")
            #wFile.write("OritatamiSystemDef=shape-definition.orit\n")
            wFile.write(assemblyString + "\n")
            #wFile.write(lastBeadString + "\n")
            wFile.close()
            globalApp.SetOutputMessage("Saved shape definition to file '" + fileName + "'")
            return True
        except:
            globalApp.ShowErrorDialog("\nERROR: Unable to write shape definition file: '" + fileName + "'\n")
            return False

    def DoSetSeed(self):
        dlg = wx.MessageDialog(self, "Are you sure you want to set the current assembly as the seed assembly?", caption="Confirm seed setting")
        if dlg.ShowModal() == wx.ID_OK:
            dlg.Destroy()
            assemblyList = self.GetAssemblyList()
            if len(assemblyList) == 0:
                globalApp.ShowErrorDialog("The seed assembly is empty.  It must contain at least one bead.")
                return False
            self.onodegrid.seed = assemblyList
            self.onodegrid.lastSeedLoc = self.onodegrid.lastBeadLoc
        else:
            dlg.Destroy()

    #  Format for assembly file:
    #  [[<bead location>], [<bead location>], ...]
    #  Format for bead location:
    #  ["<type name>", (<coords>), (<parent coords>), [<list of coords bonded to>]]
    def DoSaveAssembly(self, bOnlySeed=False, bSaveAs=False):
        aStr = "assembly"
        AStr = "Assembly"
        fileName = globalApp.assemblyDefinitionFile
        if bOnlySeed == True:
            aStr = "seed"
            AStr = "Seed"
            fileName = globalApp.seedDefinitionFile
        if fileName is None or bSaveAs == True:
            saveFileDialog = wx.FileDialog(self, "Save " + AStr + " As", "", "", 
                                           "Oritatami " + aStr + " files (*.oass)|*.oass",
                                           wx.FD_SAVE | wx.FD_OVERWRITE_PROMPT)
            if not saveFileDialog.ShowModal() == wx.ID_OK:
                return False
            fileName = saveFileDialog.GetPath()
            saveFileDialog.Destroy()

        if globalApp.systemDefinitionFile is None or len(globalApp.systemDefinitionFile) < 1:
            globalApp.ShowErrorDialog("\nERROR: In order to save an assembly, you must first have a valid oritatami system definition file associated with it. Please save your system definition first.\n")
            return False

        systemDefString = "OritatamiSystemDef=" + globalApp.systemDefinitionFile
        assemblyString = "Assembly="
        if bOnlySeed == True:
            lastBeadString = "LastBead=" + str(self.onodegrid.lastSeedLoc)
        else:
            lastBeadString = "LastBead=" + str(self.onodegrid.lastBeadLoc)
        assemblyList = self.GetAssemblyList(bOnlySeed)
        if len(assemblyList) == 0:
            globalApp.ShowErrorDialog("The seed assembly is empty.  It must contain at least one bead.")
            return False
        assemblyString = assemblyString + str(assemblyList)
        
        origFileName = fileName
        if not fileName is None:
            if len(fileName) < 1:
                return False
            elif os.path.isabs(fileName) == False:
                fileName = "./" + fileName
                fileName = os.path.abspath(fileName)

            try:
                wFile = open(fileName, "w")
                wFile.write(systemDefString + "\n")
                wFile.write(assemblyString + "\n")
                wFile.write(lastBeadString)
                wFile.close()
                if bOnlySeed == True:
                    globalApp.seedDefinitionFile = origFileName
                else:
                    globalApp.assemblyDefinitionFile = origFileName
                globalApp.SetOutputMessage("Saved oritatami " + aStr + " to file '" + fileName + "'")
                return True
            except:
                globalApp.ShowErrorDialog("\nERROR: Unable to write oritatami " + aStr + " to file: '" + fileName + "'\n")
                return False

    def GetAssemblyList(self, bOnlySeed=False):
        assemblyList = []
        if bOnlySeed == True:
            assemblyList = self.onodegrid.seed
        else:
            for i in range(self.onodegrid.minX, self.onodegrid.maxX+1):
                for j in range(self.onodegrid.minY, self.onodegrid.maxY+1):
                    locList = self.GetLocOutputList((i,j))
                    if not locList is None:
                        assemblyList.append(locList)
        return assemblyList

    def GetLocOutputList(self, loc):
        onode = self.onodegrid.GetLoc(loc)
        if onode is not None:
            #self.onodegrid.ONodeTypes[onode.type]
            locList = []
            locList.append(onode.type)
            locList.append(loc)
            if onode.parent is None:
                locList.append(None)
            else:
                locList.append(onode.parent.coords)
            locList.append(onode._bonds)
            return locList
        else:
            return None
        
    def DoSaveAll(self):
        
        if self.DoSaveShapeDef(bSaveAs=False) == True:
            #if self.DoSaveAssembly(bOnlySeed=True, bSaveAs=False) == True:
            return True
        return False

    def SetRandomSeed(self, seed):
        self.randSeed = seed
        if self.randSeed < 0:
            self.randSeed = random.randint(0, sys.maxint)
        print("Seeding random number generator with " + str(self.randSeed))
        random.seed(self.randSeed)

    def DoNew(self):
        self.onodegrid = ONodeGrid(GRIDSTARTSIZE, GRIDSTARTSIZE)

        self.SetRandomSeed(self.randSeed)
        globalApp.seedDefinitionFile = None
        globalApp.assemblyDefinitionFile = None
        globalApp.systemDefinitionFile = None
        globalApp.systemDefinitionFileFullPath = None
        
        self.drawingModeBeadType = ShapeNodeType
        self.InitializeDrawing(self.pdc)
        self.drawingModeBeadType = ShapeNodeType
        self.Refresh()
        self.ScrollToCoords((0,0))

        return False

    def Reset(self):
        #dlg = wx.MessageDialog(self, "Are you sure you want to reset your system to the\nunscaled and rotated shape definition?", caption="Confirm reset", style=wx.OK|wx.CANCEL)
        #if not dlg.ShowModal() == wx.ID_OK:
        #    return
        #dlg.Destroy()
        self.onodegrid.ResetToShapeDef()
        self.InitializeDrawing(self.pdc)
        self.Refresh()

    def Reload(self):
        dlg = wx.TextEntryDialog(self, "Are you sure you want to reload your system\nand abandon the current assembly?\nIf so, you can set your random seed below\n(use -1 to choose a new random value):", defaultValue=str(self.randSeed))
        if not dlg.ShowModal() == wx.ID_OK:
            return
        try:
            result = int(dlg.GetValue())
            self.SetRandomSeed(result)
        except ValueError:
            globalApp.ShowErrorDialog("Random seed value must be an integer.")
            dlg.Destroy()
            return
        dlg.Destroy()

        self.onodegrid = ONodeGrid(GRIDSTARTSIZE, GRIDSTARTSIZE)
        if not globalApp.seedDefinitionFile is None:
            self.onodegrid.LoadOritatamiAssembly(globalApp.seedDefinitionFile, True)
        elif not globalApp.systemDefinitionFile is None:
            self.onodegrid.LoadOritatamiSystem(globalApp.systemDefinitionFile, globalApp.systemDefinitionFileFullPath)  
        self.InitializeDrawing(self.pdc)
        self.Refresh()
        
    def Compile(self):

        # Check for connectivity
        self.app.SetOutputMessage("Checking connectivity of shape...")

        orderedList = self.onodegrid.GetOrderedList()
        print("Found " + str(len(orderedList)) + " points: " + str(orderedList))
        self.onodegrid.orderedList = orderedList
        self.onodegrid.scaledAndRotatedList = []

        if len(orderedList) < self.onodegrid._numBeads:
            self.app.SetOutputMessage("The shape is not connected!  Aborting.")
            return False
        self.app.SetOutputMessage("The shape is connected.  Continuing...")
        
        # Rotate and scale
        topLeft = orderedList[0]
        xb = topLeft[0]
        yb = topLeft[1]
        scaledAndRotated = [topLeft]

        for p in orderedList[1:]:
            x = p[0]
            y = p[1]
            
            xDiff = x - xb
            yDiff = y - yb
            xs = (2 * xDiff) + yDiff
            ys = yDiff - xDiff
            
            newPt = (xb + xs, yb + ys)
            #print(str(p) + " -> " + str(newPt))
            scaledAndRotated.append(newPt)
        
        savedShapePoints = self.onodegrid.shapePoints
        self.onodegrid = ONodeGrid(GRIDSTARTSIZE, GRIDSTARTSIZE)
        self.onodegrid.shapePoints = savedShapePoints
        self.onodegrid.orderedList = orderedList
        self.onodegrid.scaledAndRotatedList = scaledAndRotated
        self.onodegrid.gadgetSet = GadgetSet()
        self.onodegrid.currentGadget = 0

        for p in scaledAndRotated:
            self.onodegrid.seed.append([ScaledNodeType.name, p])
            loc = ONode("scaledNode", p)
            self.onodegrid.SetLoc(p, loc)

        #self.onodegrid.ResetToShapeDef()
        self.app.SetOutputMessage("Shape has been successfully scaled and rotated.")
        self.InitializeDrawing(self.pdc)
        self.Refresh()
        
    def GenerateOritatamiSystem(self):

        # Generate the Hamiltonian path
        allGadgets = self.onodegrid.GetAllGadgets()
        self.app.SetOutputMessage("There are " + str(len(allGadgets)) + " gadgets: " + str(allGadgets))
        
        allNodes = []
        for gCoords in allGadgets:
            allNodes.append(gCoords)
            for n in ONBRS:
                coords = (gCoords[0]+n[0], gCoords[1]+n[1])
                if not coords in allNodes:
                    allNodes.append(coords)
        self.app.SetOutputMessage("There are " + str(len(allNodes)) + " nodes: " + str(allNodes))
        
        visitedNodes = []
        currentNode = self.onodegrid.scaledAndRotatedList[0]
        firstNode = currentNode
        bFirst = True
        
        while bFirst == True or not currentNode == firstNode:
            bFirst = False
            #print("currentNode: " + str(currentNode))
            visitedNodes.append(currentNode)

            nbrGadgets = []
            for n in ONBRS:
                pn = (currentNode[0]+n[0], currentNode[1]+n[1])
                g = self.onodegrid.gadgetSet.Get(pn)
                if not g is None:
                    nbrGadgets.append(g)

            nextNode = None
            for gadget in nbrGadgets:
                for e in gadget.edges:
                    loc1 = (gadget.coords[0]+e[0][0], gadget.coords[1]+e[0][1])
                    loc2 = (gadget.coords[0]+e[1][0], gadget.coords[1]+e[1][1])
                    if loc2 == currentNode and not loc1 in visitedNodes:
                        nextNode = loc1
                        break
                    elif loc1 == currentNode and not loc2 in visitedNodes:
                        nextNode = loc2
                        break
            if nextNode == None:
                #print("All done!")
                break
            currentNode = nextNode
        
        numVis = len(visitedNodes)
        if numVis == len(allNodes):
            self.app.SetOutputMessage("The Hamiltonian path hit all " + str(numVis) + " nodes: " + str(visitedNodes))
        else:
            self.app.SetOutputMessage("ERROR: The attempted Hamiltonian path hit only " + str(numVis) + " out of " + str(len(allNodes)) + " nodes: " + str(visitedNodes))

        # Generate the Oritatami system
        gadgetSequence = []
        i = 0
        
        while i < numVis:
            # Get next set of nodes
            n1 = visitedNodes[i]
            n2 = visitedNodes[(i+1)%numVis] # assumes visitedNodes to be a cycle
            n3 = visitedNodes[(i+2)%numVis]

            # Get offsets from middle then compare distance
            diff1 = (n2[0]-n1[0], n2[1]-n1[1])
            diff3 = (n2[0]-n3[0], n2[1]-n3[1])
            nbr1 = ONBRINDEX[diff1]
            nbr3 = ONBRINDEX[diff3]
            angle = (nbr3 - nbr1) % 6
            #print("Angle " + str(n1) + "->" + str(n2) + "->" + str(n3) + " is: " + str(angle))
            gadgetSequence.append(angle)

            i = i + 1
            
        # Get desired output file name from user
        saveFileDialog = wx.FileDialog(self, "Save Oritatami System As", "", "", 
                                        "Oritatami assembly files (*.oass)|*.oass",
                                        wx.FD_SAVE | wx.FD_OVERWRITE_PROMPT)
        if not saveFileDialog.ShowModal() == wx.ID_OK:
            globalApp.ShowErrorDialog("You must provide a file name to save the Oritatami system to.")
            return False
        fileName = saveFileDialog.GetPath()
        saveFileDialog.Destroy()
        
        if (not len(fileName) > len(".oass")) or (not fileName[-5:] == ".oass"):
            fileName = fileName + ".oass"
        #if os.path.isabs(fileName):
        # os.path.realpath(__file__)  will get path to executing script
        #    fileName = os.path.relpath(fileName)
        #    print("basename: " + str(os.path.basename(fileName)))
        oassFullFileName = fileName
        oritFullFileName = fileName[:-5] + ".orit"
        oritFileName = os.path.basename(oritFullFileName)
        
        #print("Going to save to " + oassFullFileName + " and " + oritFullFileName + " orit base: " + oritFileName)
        
        # Load the definition files to use as a template

        os.stat(DEFINITION_FILE_OASS)
        rFile = open(DEFINITION_FILE_OASS, "r")
        print("Using Oritatami system definition template '" + DEFINITION_FILE_OASS + "'")
        lines = rFile.readlines()
        rFile.close()
        seedLine = ""
        lastBeadLine = ""
        systemDefLine = "OritatamiSystemDef=" + oritFileName + "\n"
        for line in lines:
            if line.startswith("Assembly="):
                seedLine = line
            elif line.startswith("LastBead="):
                lastBeadLine = line
        
        os.stat(DEFINITION_FILE_ORIT)
        rFile = open(DEFINITION_FILE_ORIT, "r")
        print("Using Oritatami seed definition template '" + DEFINITION_FILE_ORIT + "'")
        lines = rFile.readlines()
        rFile.close()
        beadTypeLine = ""
        delayLengthLine = ""
        arityLine = ""
        lwbLine = ""
        bffsLine = ""
        for line in lines:
            if line.startswith("BeadTypes="):
                beadTypeLine = line
            elif line.startswith("DelayLength="):
                delayLengthLine = line
            elif line.startswith("Arity="):
                arityLine = line
            elif line.startswith("LockWithoutBinding="):
                lwbLine = line
            elif line.startswith("BindFullFloppySegment="):
                bffsLine = line
        beadCycleLine = "BeadCycle=["

        os.stat(DEFINITION_FILE_TXT)
        rFile = open(DEFINITION_FILE_TXT, "r")
        print("Using Oritatami gadget definition template '" + DEFINITION_FILE_TXT + "'")
        lines = rFile.readlines()
        rFile.close()
        gadgetDict = {}
        for line in lines:
            l = line.split("=")
            key = l[0]
            seq = l[1].strip()
            gadgetDict[key] = seq
        
        bFirst = True
        last = None
        
        # the first table below represent colors that are even (b) or odd (g) steps away from a red 'r' tile with previous tile blue 'b'
        # other table entries are similar 
        colorTable      = {}
        colorTable[('r','b')] = ['b','g']
        colorTable[('r','g')] = ['g','b']
        colorTable[('g','b')] = ['b','r']
        colorTable[('g','r')] = ['r','b']
        colorTable[('b','g')] = ['g','r']
        colorTable[('b','r')] = ['r','g']
        
        lastColor = None
        lastLastColor = None
        
        
        # TODO: seed to get the correct rotation of the shape. The following definition hard codes the seed. 
        seedLine=Assembly=[['1r', (0, 0), None, []], ['2r', (1, 0), (0, 0), [(1, 1)]], ['3r', (2, 0), (1, 0), []], ['4r', (2, -1), (2, 0), [(1, 0)]]]
        
        
        
        for g in gadgetSequence:
            if bFirst == False:
                beadCycleLine = beadCycleLine + ","
                
            #print("Appending gadget " + str(g))
            if bFirst == True:
                
                # create the 4 bead seed line
                firstGadget = gadgetDict[str(g) + 'r'].split(",")
                for i in range(4):
                    seedLine[i][0] = firstGadget[i].replace("'","")
                seedLine = "Assembly=" + str(seedLine) + "\n"

                # The first gadget in the transcript is "missing" 4 beads, which are part of the seed
                beadCycleLine = beadCycleLine + ",".join(firstGadget[4:])
                
                
                lastColor = 'r'
                lastLastColor = 'g'
                bFirst = False
            else:                        
                currentColor = colorTable[(lastColor,lastLastColor)][last % 2]
                beadCycleLine = beadCycleLine + gadgetDict[str(g) + currentColor]
                lastLastColor = lastColor
                lastColor = currentColor
                
            last = g
            
        #print("Appending final gadget " + str(1))
        # beadCycleLine = beadCycleLine + "," + gadgetDict[str(1) + colorTable[(lastColor,lastLastColor)][last % 2]]

        beadCycleLine = beadCycleLine + "]\n"
        
        #print len(ast.literal_eval(beadCycleLine.split("=")[1]))
        #print("beadCycleLine: " + beadCycleLine)
        

        wFile = open(oassFullFileName, "w")
        wFile.write(systemDefLine)
        wFile.write(seedLine)
        wFile.write(lastBeadLine)
        wFile.close()
        
        wFile = open(oritFullFileName, "w")
        wFile.write(beadTypeLine)
        wFile.write(beadCycleLine)
        wFile.write(delayLengthLine)
        wFile.write(arityLine)
        wFile.write(lwbLine)
        wFile.write(bffsLine)
        wFile.close()

        self.app.SetOutputMessage("Saved to output files " + oritFullFileName + " and " + oassFullFileName)

    def AddGadget(self):

        if self.onodegrid.scaledAndRotatedList is None or len(self.onodegrid.scaledAndRotatedList) == 0:
            self.app.SetOutputMessage("Unable to add a gadget.  A connected shape must first be scaled and rotated.")
            return False

        if self.onodegrid.currentGadget >= len(self.onodegrid.scaledAndRotatedList):
            self.app.SetOutputMessage("There are no more points to replace with gadgets.")
            return False
        
        # Replace next scaled point with a gadget
        p = self.onodegrid.scaledAndRotatedList[self.onodegrid.currentGadget]
        self.onodegrid.currentGadget = self.onodegrid.currentGadget + 1
        ng = NodeGadget(p)
        self.onodegrid.gadgetSet.Add(ng)

        abstractNbrPattern = self.onodegrid.gadgetSet.GetAbstractNbrPattern(p)
        #print("Got abstractNbrPattern " + str(abstractNbrPattern))
        (matchedPattern, rot, reflect, changeList) = MatchPattern(abstractNbrPattern)
        
        if not matchedPattern is None:
            if self.onodegrid.gadgetSet.ExtendCycle(ng.coords, rot, reflect, changeList) == True:
                self.InitializeDrawing(self.pdc)
                self.Refresh()
                return True
        else:
            self.app.SetOutputMessage("ERROR:  Could not find matching pattern for neighboring gadgets!")
            return False


    def DoForwardStep(self):

        return self.AddGadget()
        
        #MJP TODO: Switch back to efficient drawing later.  For now, just redraw everything for each step!
        #self.InitializeDrawing(self.pdc)
        #self.ScrollToCoords(coords)

    def DoGo(self):
        bContinue = self.DoForwardStep()
        while bContinue == True:
            bContinue = self.DoForwardStep()

    def ZoomThenRecenter(self, zoomIn=True):
        usp = self.CalcUnscrolledPosition(0,0)
        sz = self.GetSize()
        center = (usp[0] + sz[0]/2, usp[1] + sz[1]/2)
        coords = self.GetCoordsFromPixels(center[0], center[1], True)
        self.DoScale(zoomIn)
        self.ScrollToCoords(coords)
    
    def ConvertEventCoords(self, event):
        xView, yView = self.GetViewStart()
        xDelta, yDelta = self.GetScrollPixelsPerUnit()
        return (event.GetX() + (xView * xDelta),
            event.GetY() + (yView * yDelta))

    def OffsetRect(self, r):
        xView, yView = self.GetViewStart()
        xDelta, yDelta = self.GetScrollPixelsPerUnit()
        r.OffsetXY(-(xView*xDelta),-(yView*yDelta))

    def OnChar(self, evt):
        keycode = evt.GetKeyCode()
        if keycode == ord("a") or keycode == ord("A"):
            self.AddTileCopiesToDrawing(self.pdc)
            self.Refresh()
        elif keycode == ord("p") or keycode == ord("P"):
            self.PrintDebugInfo()
        elif keycode == ord("r") or keycode == ord("R"):
            self.Reset()
        elif keycode == ord("s") or keycode == ord("S"):
            self.DoSaveAll()
        elif keycode == ord("h") or keycode == ord("H"):
            self.ShowHelp()
        elif keycode == ord("l") or keycode == ord("L"):
            self.DoLoad(bAssembly=True, bSystem=True)
        elif keycode == ord("f") or keycode == ord("F"):
            self.DoForwardStep()
        elif keycode == ord("g") or keycode == ord("G"):
            self.DoGo()
        elif keycode == ord("=") or keycode == ord("+"):
            self.DoScale(True)
        elif keycode == ord("-") or keycode == ord("_"):
            self.DoScale(False)
        elif keycode == ord("c") or keycode == ord("C"):
            self.ScrollToCoords((0,0))
        elif keycode == ord("t") or keycode == ord("T"):
            self.ToggleGrid()
        elif keycode == ord("m") or keycode == ord("M"):
            self.MessageToggle(not self.app.showMessages)
            self.app.menuView.Check(self.app.itemTM.GetId(), self.app.showMessages)
        elif keycode == ord("d") or keycode == ord("D"):
            self.DisplayToggle(not self.app.showDisplayUpdates)
            self.app.menuView.Check(self.app.itemTD.GetId(), self.app.showDisplayUpdates)
        elif keycode == ord("b") or keycode == ord("B"):
            self.SetBreakpoints()
        #elif keycode == wx.WXK_DELETE or keycode == wx.WXK_BACK:
            #self.DoDelete()

    def ShowHelp(self):
        self.app.OnHelp(None)

    def PrintDebugInfo(self):        
        #self.app.SetOutputMessage("some stuff")
        return

    def DoScale(self, bIncrease):
        #wx.MessageBox("Oops, this doesn't really work yet...", "Zooming not supported", wx.ICON_EXCLAMATION)
        #return
        if bIncrease == True:
            self.scale = self.scale * 2.0
        else:
            self.scale = self.scale / 2.0
        print("Scale set to " + str(self.scale))
        
        self.SetVirtualSize((self.winWidth*self.scale, self.winHeight*self.scale))

        self.Refresh()


    def OnMouse(self, event):
        
        #=======================================================================
        # if event.IsPageScroll():
        #     print("is page scroll")
        # if event.MiddleIsDown():
        #     print("middle is down")
        #     if event.Dragging():
        #         print("Dragging")
        #=======================================================================
            
        x,y = self.ConvertEventCoords(event)
        
        if self.showMouseOver == True:
            x2,y2 = self.GetCoordsFromPixels(x, y)
    
            if x2 == None:
                self.frame.SetStatusText("", 0)
                self.frame.SetStatusText("", 1)
            else:
                self.frame.SetStatusText("Coords: (" + str(int(x2)) + "," + str(int(y2)) + ")", 0)
                onode = self.onodegrid.GetLoc((x2,y2))
                if not onode is None:
                    outStr = "Bead type: " + str(onode.type)
                    self.frame.SetStatusText(outStr, 1)
                else:
                    self.frame.SetStatusText("", 1)
        
        if event.LeftDown():
            if self.mouseMode == ORIT_MODE_PAN_ZOOM:
                self.SetCursor(wx.CursorFromImage(self.grabbingCursorImg))
                self.dragStart = self.ScreenToClient(wx.GetMousePosition())
                self.dragStartScrollPos = (self.GetScrollPos(wx.HORIZONTAL), self.GetScrollPos(wx.VERTICAL))

            elif self.mouseMode == ORIT_MODE_DRAW_BEAD:
                if self.drawingModeBeadType is None:
                    globalApp.ShowErrorDialog("A bead type must be selected in the bead type editor window")
                else:
                    x2,y2 = self.GetCoordsFromPixels(x, y)
                    if (x2 is not None) and (y2 is not None):
                        if self.onodegrid.GetLoc((x2, y2)) is None:
                            parentLoc = None
                            if (self.onodegrid.lastBeadLoc is not None) and self.onodegrid.AreNbrLocs((x2, y2), self.onodegrid.lastBeadLoc) == True:
                                parentLoc = self.onodegrid.lastBeadLoc
                            self.ManuallyAddNode(self.drawingModeBeadType.name, (x2, y2), parentLoc)
            elif (self.mouseMode == ORIT_MODE_DRAW_LINK) or (self.mouseMode == ORIT_MODE_DRAW_BOND):
                if (self.linkStart is None) or (self.negSelected is not None):
                    xs,ys = self.GetCoordsFromPixels(x, y)
                    if (xs is None) or (ys is None):
                        self.linkStart = None
                    else:
                        self.linkStart = (xs,ys)
                        node = self.onodegrid.GetLoc(self.linkStart)
                        if self.negSelected is not None:
                            saved = self.negSelected.coords
                            self.negSelected = None
                            self.ManuallyDrawNodes([saved])
                        self.posSelected = node
                        self.ManuallyDrawNodes([self.linkStart])
                else:
                    xe,ye = self.GetCoordsFromPixels(x, y)
                    if self.linkStart[0] == xe and self.linkStart[1] == ye:
                        node = self.onodegrid.GetLoc(self.linkStart)
                        self.posSelected = None
                        self.ManuallyDrawNodes([self.linkStart])
                        self.linkStart = None
                    else:
                        if (xe is None) or (ye is None):
                            self.linkEnd = None
                        else:
                            self.linkEnd = (xe,ye)
                        if (self.linkStart is not None) and (self.linkEnd is not None):
                            parentNode = self.onodegrid.GetLoc(self.linkStart)
                            childNode = self.onodegrid.GetLoc(self.linkEnd)
                            if (parentNode is not None) and (childNode is not None):
                                if self.mouseMode == ORIT_MODE_DRAW_LINK:
                                    if (self.onodegrid.AreNbrLocs(self.linkStart, self.linkEnd) == True):
                                        if parentNode.child is not None:
                                            globalApp.ShowErrorDialog("The node at " + str(self.linkStart) + " already has a child")
                                        elif childNode.parent is not None:
                                            globalApp.ShowErrorDialog("The node at " + str(self.linkEnd) + " already has a parent")
                                        else:
                                            bAlreadyBonded = False
                                            for n in range(parentNode.NumBonds()):
                                                b = parentNode.GetNthBond(n)
                                                if b == self.linkEnd:
                                                    bAlreadyBonded = True
                                            if bAlreadyBonded == True:
                                                globalApp.ShowErrorDialog("Nodes must not already be bonded to be linked as parent and child")
                                            else:
                                                parentNode.child = childNode
                                                childNode.parent = parentNode
                                                self.posSelected = None
                                                self.ManuallyDrawNodes([self.linkStart, self.linkEnd])
                                                self.linkStart = None
                                                self.linkEnd = None
                                    else:
                                        globalApp.ShowErrorDialog("Nodes must be neighbors to be linked")
                                elif self.mouseMode == ORIT_MODE_DRAW_BOND:
                                    if (self.onodegrid.AreNbrLocs(self.linkStart, self.linkEnd) == True):
                                        bAlreadyBonded = False
                                        for n in range(parentNode.NumBonds()):
                                            b = parentNode.GetNthBond(n)
                                            if b == self.linkEnd:
                                                bAlreadyBonded = True
                                        if bAlreadyBonded == True:
                                            globalApp.ShowErrorDialog("Nodes are already bonded")
                                        elif (parentNode.child == childNode) or (childNode.child == parentNode):
                                            globalApp.ShowErrorDialog("Nodes cannot be bonded if they are linked as parent and child")
                                        else:
                                            bDoIt = True
                                            numP = parentNode.NumBonds()
                                            if (self.confirmOverArity == True) and (numP >= self.onodegrid.arity):
                                                msg = "The bead at location " + str(parentNode.coords) + " currently has " + str(numP)
                                                if numP == 1:
                                                    msg = msg + " bond"
                                                else:
                                                    msg = msg + " bonds"
                                                msg = msg + ", and the arity parameter is set to " + str(self.onodegrid.arity) + ".\nDo you want to create the new bond anyway?" 
                                                dlg = DontShowAgainDialog(self, wx.ID_ANY, "Confirm exceeding arity", msg)
                                                dlg.CenterOnScreen()
                                                retId = dlg.ShowModal()
                                                dlg.Destroy()
                                                if retId == ORIT_ID_OK_NOCHECK:
                                                    self.confirmOverArity = True
                                                elif retId == ORIT_ID_OK_CHECK:
                                                    self.confirmOverArity = False
                                                else:
                                                    bDoIt = False
                                            numC = childNode.NumBonds()
                                            if (bDoIt == True) and (self.confirmOverArity == True) and (numC >= self.onodegrid.arity):
                                                msg = "The bead at location " + str(childNode.coords) + " currently has " + str(numC)
                                                if numP == 1:
                                                    msg = msg + " bond"
                                                else:
                                                    msg = msg + " bonds"
                                                msg = msg + ", and the arity parameter is set to " + str(self.onodegrid.arity) + ".\nDo you want to create the new bond anyway?" 
                                                dlg = DontShowAgainDialog(self, wx.ID_ANY, "Confirm exceeding arity", msg)
                                                dlg.CenterOnScreen()
                                                retId = dlg.ShowModal()
                                                dlg.Destroy()
                                                if retId == ORIT_ID_OK_NOCHECK:
                                                    self.confirmOverArity = True
                                                elif retId == ORIT_ID_OK_CHECK:
                                                    self.confirmOverArity = False
                                                else:
                                                    bDoIt = False
                                            if (bDoIt == True) and (self.confirmInvalidBonds == True) and (self.onodegrid.ONodeTypes[parentNode.type].BondsToNodeType(childNode.type) == False):
                                                msg = "Bead types '" + str(parentNode.type) + "' and '" + str(childNode.type) + "' don't form bonds."
                                                msg = msg + "\nDo you want to create the new bond anyway?" 
                                                dlg = DontShowAgainDialog(self, wx.ID_ANY, "Confirm bond between non-bonding types", msg)
                                                dlg.CenterOnScreen()
                                                retId = dlg.ShowModal()
                                                dlg.Destroy()
                                                if retId == ORIT_ID_OK_NOCHECK:
                                                    self.confirmInvalidBonds = True
                                                elif retId == ORIT_ID_OK_CHECK:
                                                    self.confirmInvalidBonds = False
                                                else:
                                                    bDoIt = False
                                            if bDoIt == True:
                                                self.onodegrid.AddBond(self.linkStart, self.linkEnd)
                                                self.posSelected = None
                                                self.ManuallyDrawNodes([self.linkStart, self.linkEnd])
                                                self.linkStart = None
                                                self.linkEnd = None
                                    else:
                                        globalApp.ShowErrorDialog("Nodes must be neighbors to be bonded")

        elif event.RightDown():
            if self.mouseMode == ORIT_MODE_PAN_ZOOM:
                #self.SetCursor(wx.CursorFromImage(self.grabbingCursorImg))
                doScale = True
                if (event.AltDown() == True) or (event.ControlDown() == True) or (event.ShiftDown() == True):
                    doScale = False
                x2,y2 = self.GetCoordsFromPixels(x, y, True)
                self.DoScale(doScale)
                self.ScrollToCoords((x2,y2))
                sz = self.GetSize()
                self.WarpPointer(sz[0]/2, sz[1]/2)
                #self.SetCursor(wx.StockCursor(wx.CURSOR_ARROW))

            elif self.mouseMode == ORIT_MODE_DRAW_BEAD:
                x2,y2 = self.GetCoordsFromPixels(x, y)
                if (x2 is not None) and (y2 is not None):
                    if self.onodegrid.GetLoc((x2, y2)) is not None:
                        self.ManuallyRemoveNode((x2, y2))
                        
            elif (self.mouseMode == ORIT_MODE_DRAW_LINK) or (self.mouseMode == ORIT_MODE_DRAW_BOND):
                if (self.linkStart is None) or (self.posSelected is not None):
                    xs,ys = self.GetCoordsFromPixels(x, y)
                    if (xs is None) or (ys is None):
                        self.linkStart = None
                    else:
                        self.linkStart = (xs,ys)
                        node = self.onodegrid.GetLoc(self.linkStart)
                        if self.posSelected is not None:
                            saved = self.posSelected.coords
                            self.posSelected = None
                            self.ManuallyDrawNodes([saved])
                        self.negSelected = node
                        self.ManuallyDrawNodes([self.linkStart])
                else:
                    xe,ye = self.GetCoordsFromPixels(x, y)
                    if self.linkStart[0] == xe and self.linkStart[1] == ye:
                        node = self.onodegrid.GetLoc(self.linkStart)
                        self.negSelected = None
                        self.ManuallyDrawNodes([self.linkStart])
                        self.linkStart = None
                    else:
                        if (xe is None) or (ye is None):
                            self.linkEnd = None
                        else:
                            self.linkEnd = (xe,ye)
                        if (self.linkStart is not None) and (self.linkEnd is not None):
                            parentNode = self.onodegrid.GetLoc(self.linkStart)
                            childNode = self.onodegrid.GetLoc(self.linkEnd)
                            if (parentNode is not None) and (childNode is not None):
                                bDidIt = False
                                if self.mouseMode == ORIT_MODE_DRAW_LINK:
                                    if (parentNode.child == childNode):
                                        parentNode.child = None
                                        childNode.parent = None
                                        bDidIt = True
                                    elif (childNode.child == parentNode):
                                        childNode.child = None
                                        parentNode.parent = None
                                        bDidIt = True
                                    if bDidIt == True:
                                        self.negSelected = None
                                        self.linkStart = None
                                        self.linkEnd = None
                                        self.InitializeDrawing(self.pdc)
                                    else:
                                        globalApp.ShowErrorDialog("Cannot remove link between beads which aren't linked.")
                                elif self.mouseMode == ORIT_MODE_DRAW_BOND:
                                    bAlreadyBonded = False
                                    for n in range(parentNode.NumBonds()):
                                        b = parentNode.GetNthBond(n)
                                        if b == self.linkEnd:
                                            bAlreadyBonded = True
                                    if bAlreadyBonded == True:
                                        self.onodegrid.RemoveBond(self.linkStart, self.linkEnd)
                                        bDidIt = True
                                    if bDidIt == True:
                                        self.negSelected = None
                                        self.linkStart = None
                                        self.linkEnd = None
                                        self.InitializeDrawing(self.pdc)
                                    else:
                                        globalApp.ShowErrorDialog("Cannot remove bond between beads which don't have a bond.")

        if event.LeftUp():
            self.dragStart = None
            self.SetCursor(wx.StockCursor(wx.CURSOR_ARROW))

        elif event.Dragging():
            if self.mouseMode == ORIT_MODE_PAN_ZOOM:
                if self.dragStart is not None:
                    #self.SetScrollRate(1,1)
                    new = self.ScreenToClient(wx.GetMousePosition())
                    dx = int((self.dragStart.x-new.x)/1)
                    dy = int((self.dragStart.y-new.y)/1)
                    spx = self.GetScrollPos(wx.HORIZONTAL)
                    spy = self.GetScrollPos(wx.VERTICAL)
                    self.Scroll(spx+dx, spy+dy)
                    self.dragStart = new
            
            elif self.mouseMode == ORIT_MODE_DRAW_BEAD:
                if event.LeftIsDown():
                    if self.drawingModeBeadType is None:
                        globalApp.ShowErrorDialog("A bead type must be selected in the bead type editor window")
                    else:
                        x2,y2 = self.GetCoordsFromPixels(x, y)
                        if (x2 is not None) and (y2 is not None):
                            if self.onodegrid.GetLoc((x2, y2)) is None:
                                parentLoc = None
                                if (self.onodegrid.lastBeadLoc is not None) and self.onodegrid.AreNbrLocs((x2, y2), self.onodegrid.lastBeadLoc) == True:
                                    parentLoc = self.onodegrid.lastBeadLoc
                                self.ManuallyAddNode(self.drawingModeBeadType.name, (x2, y2), parentLoc)
                elif event.RightIsDown():
                    x2,y2 = self.GetCoordsFromPixels(x, y)
                    if (x2 is not None) and (y2 is not None):
                        if self.onodegrid.GetLoc((x2, y2)) is not None:
                            self.ManuallyRemoveNode((x2, y2))

    def ManuallyAddNode(self, typeName, coords, parentCoords):
        self.onodegrid.AddNode(self.drawingModeBeadType.name, coords, parentCoords)
        self.ManuallyDrawNodes([coords])
        self.frame.SetStatusText("Number of beads: " + str(self.onodegrid.GetNumBeads()), 2)
        
    def ManuallyRemoveNode(self, coords):
        self.onodegrid.RemoveNode(coords)
        self.InitializeDrawing(self.pdc)
        self.frame.SetStatusText("Number of beads: " + str(self.onodegrid.GetNumBeads()), 2)
        
    def ManuallyDrawNodes(self, coordList):
        self._lock.acquire()
        for coords in coordList:
            self.DrawQueue.append(coords)
        self._lock.release()
        self.DrawUpdate(self.pdc)
        self.Refresh()

    def ScrollToCoords(self, coords):
        if (not coords == None) and (not coords[0] == None) and (not coords[1] == None):
            (x,y) = self.GetPixelsForCoords(coords[0], coords[1])
            x = x * self.scale
            y = y * self.scale
            (xs, ys) = self.GetScrollPixelsPerUnit()
            (xsz,ysz) = self.GetSizeTuple()
            self.Scroll((x-xsz/2+self.sideLen*self.scale)/xs, (y-ysz/2+self.sideLen*self.scale)/ys)


    def ColoredPen(self, color, width = 2):
        if not self.pen_cache.has_key( (color, width) ):
            self.pen_cache[(color, width)] = wx.Pen(color, width)
        return self.pen_cache[(color, width)]

    def ColoredBrush(self, colorName):
        if not self.brush_cache.has_key(colorName):
            if type(colorName) is str:
                if "(" in colorName:
                    colorTuple = eval(colorName)
                    if len(colorTuple) > 3:
                        color = wx.Colour(colorTuple[0], colorTuple[1], colorTuple[2], colorTuple[3])
                    else:
                        color = wx.Colour(colorTuple[0], colorTuple[1], colorTuple[2])
                else:
                    color = wx.TheColourDatabase.Find(colorName)
                self.brush_cache[colorName] = wx.Brush(color)
        return self.brush_cache[colorName]


    def OnPaint(self, event):
        #print("In OnPaint")
        # Create a buffered paint DC.  It will create the real
        # wx.PaintDC and then blit the bitmap to it when dc is
        # deleted.  
        dc = wx.BufferedPaintDC(self)
        # use PrepateDC to set position correctly
        self.PrepareDC(dc)
        # we need to clear the dc BEFORE calling PrepareDC
        bg = wx.Brush(self.GetBackgroundColour())
        dc.SetBackground(bg)
        dc.Clear()

        #dc.SetUserScale(self.scale, self.scale)
        #self.InitializeDrawing(dc)

        # create a clipping rect from our position and size
        # and the Update Region
        xv, yv = self.GetViewStart()
        dx, dy = self.GetScrollPixelsPerUnit()
        x, y   = (xv * dx, yv * dy)
        rgn = self.GetUpdateRegion()
        rgn.Offset(x,y)
        r = rgn.GetBox()
        # draw to the dc using the calculated clipping rect
        self.pdc.DrawToDCClipped(dc,r)



    def DrawUpdate(self, dc):
       
        dc.BeginDrawing()
        
        # Thread-safely make a copy of the list to draw then clear it out
        self._lock.acquire()
        copiedList = list(self.DrawQueue)
        if self.lastCurrent is not None:    # Redraw the bead that was current at the last step
            copiedList.append(self.lastCurrent)
        self.DrawQueue = []
        self._lock.release()
        
        for e in copiedList:
            onode = self.onodegrid.GetLoc((e[0], e[1]))
            if onode is not None:
                self.DrawAtCoords(dc, e[0], e[1], onode, self.onodegrid)

        #=======================================================================
        # for i in range(self.onodegrid.minX, self.onodegrid.maxX+1):
        #     for j in range(self.onodegrid.minY, self.onodegrid.maxY+1):
        #         onode = self.onodegrid.GetLoc((i,j))
        #         if onode is not None:
        #             self.DrawAtCoords(dc, i, j, onode, self.onodegrid)
        #=======================================================================

        if len(self.RemoveQueue) > 0:
            self._lock.acquire()
            self.InitializeDrawing(self.pdc)
            self.RemoveQueue = []
            self._lock.release()
            
        for e in self.RemoveQueue:
            dc.RemoveId(e)
            #pen = self.ColoredPen("BLACK")
            #dc.SetPen(pen)
            #dc.SetBrush(self.ColoredBrush("WHITE"))
            #dc.DrawRectangle(0,0,300,300)                            
        self.RemoveQueue = []
        dc.EndDrawing()
        
    def InitializeDrawing(self, dc):
       
        # Compute new dimensions
        self.gridWidth = self.onodegrid.GetWidth()
        self.gridHeight = self.onodegrid.GetHeight()
        
        self.winWidth  = self.sideLen * self.gridWidth  + self.GetHorizOffset(self.gridWidth)
        self.winHeight = self.GetVertOffset(self.gridHeight)

        print("In InitializeDrawing with scale " + str(self.scale))
        self.SetVirtualSize((self.winWidth*self.scale, self.winHeight*self.scale))

        #  Remove old data
        dc.RemoveAll()
        dc.Clear()
        #self.Refresh()

        dc.BeginDrawing()
        
        for i in range(self.onodegrid.minX, self.onodegrid.maxX+1):
            for j in range(self.onodegrid.minY, self.onodegrid.maxY+1):
                onode = self.onodegrid.GetLoc((i,j))
                if onode is not None:
                    if onode.type == "scaledNode":
                        gs = self.onodegrid.gadgetSet.Get(onode.coords)
                        if not gs is None:
                            gs.DrawPolygon(dc, self)

        dc.SetPen(self.gridPen)
        if (self.showGrid == SHOW_GRID_AND_COORDS) or (self.showGrid == SHOW_GRID_ONLY):
            # Draw horizontal lines            
            for i in range(self.onodegrid.minX, self.onodegrid.maxX+1):
                startPt = self.GetPixelsForCoords(self.onodegrid.minX, i)
                endPt = self.GetPixelsForCoords(self.onodegrid.maxX, i)
                dc.DrawLine(startPt[0], startPt[1], endPt[0], endPt[1])

            # Draw positive sloped vertical lines
            #MJP: assuming a square grid (i.e. same number vertical and horizontal cells)
            for i in range(self.onodegrid.minX, self.onodegrid.maxX+1):
                startPt = self.GetPixelsForCoords(i, self.onodegrid.minY)
                endPt = self.GetPixelsForCoords(i, self.onodegrid.maxY)
                dc.DrawLine(startPt[0], startPt[1], endPt[0], endPt[1])
                startPt = self.GetPixelsForCoords(self.onodegrid.maxX, i)
                endPt = self.GetPixelsForCoords(i, self.onodegrid.maxY)
                #dc.DrawLine(startPt[0], startPt[1], endPt[0], endPt[1])

            # Draw negative sloped vertical lines
            #MJP: assuming a square grid (i.e. same number vertical and horizontal cells)
            for i in range(self.onodegrid.minX, self.onodegrid.maxX+1):
                startPt = self.GetPixelsForCoords(i, self.onodegrid.minY)
                endPt = self.GetPixelsForCoords(self.onodegrid.minX,i)
                dc.DrawLine(startPt[0], startPt[1], endPt[0], endPt[1])
                startPt = self.GetPixelsForCoords(i, self.onodegrid.maxY)
                endPt = self.GetPixelsForCoords(self.onodegrid.maxX, i)
                dc.DrawLine(startPt[0], startPt[1], endPt[0], endPt[1])

        if self.showGrid == SHOW_GRID_AND_COORDS:
            for i in range(self.onodegrid.minX, self.onodegrid.maxX+1):
                for j in range(self.onodegrid.minY, self.onodegrid.maxY+1):
                    (xp,yp) = self.GetPixelsForCoords(i,j)
                    dc.DrawText(str(i) + "," + str(j), xp+8, yp-2)

        for i in range(self.onodegrid.minX, self.onodegrid.maxX+1):
            for j in range(self.onodegrid.minY, self.onodegrid.maxY+1):
                onode = self.onodegrid.GetLoc((i,j))
                if onode is not None:
                    if onode.type == "shapeNode":
                        self.DrawAtCoords(dc, i, j, onode, self.onodegrid)
                    elif onode.type == "scaledNode":
                        gs = self.onodegrid.gadgetSet.Get(onode.coords)
                        if not gs is None:
                            gs.Draw(dc, self)
                        else:
                            self.DrawAtCoords(dc, i, j, onode, self.onodegrid)

        dc.EndDrawing()
        self.Refresh()
        self.frame.SetStatusText("Number of beads: " + str(self.onodegrid.GetNumBeads()), 2)


    def GetHorizOffset(self, y):
        return y * self.sideLen * math.cos(math.radians(60))
    
    def GetVertOffset(self, y):
        return y * self.sideLen * math.cos(math.radians(30))

    def GetPixelsForCoords(self, xCoord, yCoord):
        xCoordAdj = xCoord - self.onodegrid.minX
        yCoordAdj = yCoord - self.onodegrid.minY
        xPixel = (xCoordAdj * self.sideLen) + self.GetHorizOffset(yCoordAdj)
        yPixel = self.GetVertOffset(yCoordAdj)
        sz = self.GetVirtualSize()
        yPixel = sz[1] - (yPixel*self.scale)
        return (xPixel*self.scale, yPixel)

    def GetCoordsFromPixels(self, xPixel, yPixel, bGapOkay=False):
        sz = self.GetVirtualSize()
        yPixel = sz[1] - yPixel
        xPixel = xPixel / self.scale
        yPixel = yPixel / self.scale
        yCoord = (yPixel / (self.sideLen * math.cos(math.radians(30)))) + self.onodegrid.minX
        xCoord = ((xPixel - self.GetHorizOffset(yCoord - self.onodegrid.minY)) / self.sideLen) + self.onodegrid.minX
        (x2,x1) = math.modf(xCoord)
        (y2,y1) = math.modf(yCoord)
        if ((abs(x2) < 0.2 or abs(x2) > 0.8) and (abs(y2) < 0.2 or abs(y2) > 0.8)) or (bGapOkay == True):
            retX = int(round(xCoord))
            retY = int(round(yCoord))
            if (retX >= self.onodegrid.minX and retX <= self.onodegrid.maxX and retY >= self.onodegrid.minY and retY <= self.onodegrid.maxY):
                return (retX, retY)
        return (None,None)

    def DrawAtCoords(self, dc, xCoord, yCoord, onode, grid):
        
        (xPixel,yPixel) = self.GetPixelsForCoords(xCoord, yCoord)

        dc.SetPen(self.gridPen)

        if onode.type in grid.ONodeTypes:
            nodeColor = grid.ONodeTypes[onode.type].color
        else:
            nodeColor = "Black"
        pointBrush = self.ColoredBrush(nodeColor)
        dc.SetBrush(pointBrush)
        dc.DrawCirclePoint((xPixel,yPixel), 10)

        if onode == self.posSelected:
            pointBrush = self.ColoredBrush("Green")
            dc.SetBrush(pointBrush)
            dc.DrawCirclePoint((xPixel,yPixel), 5)
        elif onode == self.negSelected:
            pointBrush = self.ColoredBrush("Red")
            dc.SetBrush(pointBrush)
            dc.DrawCirclePoint((xPixel,yPixel), 5)
        elif onode.coords == self.onodegrid.lastBeadLoc:
            #pointBrush = self.ColoredBrush("White")
            #dc.SetBrush(pointBrush)
            #dc.DrawCirclePoint((xPixel,yPixel), 5)
            self.lastCurrent = onode.coords
        
    def ToggleGrid(self):
            self.showGrid = (self.showGrid + 1) % 3
            self.InitializeDrawing(self.pdc)
            
    def MessageToggle(self, bShow):
        self.app.showMessages = bShow
        if bShow == True:
            for msg in self.app.messageQueue:
                self.app.SetOutputMessage(msg)
        self.app.messageQueue = []

    def DisplayToggle(self, bShow):
        self.app.showDisplayUpdates = bShow
        if self.app.showDisplayUpdates == True:
            self.DrawUpdate(self.pdc)
            self.Refresh()

#---------------------------------------------------------------------------


class OritatamiSimulatorApp(wx.App):
    def __init__(self):
        wx.App.__init__(self, redirect=False)
        self.showMessages = True
        self.messageQueue = []
        self.showDisplayUpdates = True
        self.useAdvancedColorPicker = False
        global globalApp
        globalApp = self


    def OnInit(self):

        wx.Log_SetActiveTarget(wx.LogStderr())

        self.SetAssertMode(assertMode)

        self.frame = wx.Frame(None, -1, "Oritatami Shape Maker", pos=(50,50), size=(400,400),
                         style=wx.DEFAULT_FRAME_STYLE, name="Oritatami Shape Maker")
        self.frame.SetIcon(wx.IconFromBitmap(wx.Bitmap("images/logo.png", wx.BITMAP_TYPE_PNG)))
        self.frame.CreateStatusBar(3)
        
        self.frame._mgr = wx.aui.AuiManager(self.frame)
        self.beadTypePanel          = None
        self.seqPanel               = None
        self.beadSequenceWin        = None
        self.messageWin             = None
        self.bpSize                 = (0,0)
        self.systemDefinitionFile   = None
        self.systemDefinitionFileFullPath = None
        self.seedDefinitionFile     = None
        self.assemblyDefinitionFile = None
        self.docFrame               = None
        
        provider = wx.SimpleHelpProvider()
        wx.HelpProvider.Set(provider)

        menuBar = wx.MenuBar()
        menu = wx.Menu()
        itemNew = menu.Append(-1, "New shape definition", "New shape definition")
        self.Bind(wx.EVT_MENU, self.OnNew, itemNew)
        menu.AppendSeparator()
        #itemLSys = menu.Append(-1, "Load system definition", "Load system definition")
        #self.Bind(wx.EVT_MENU, self.OnLoadSystem, itemLSys)
        item1 = menu.Append(-1, "&Load shape definition\tCtrl-L", "Load shape definition")
        self.Bind(wx.EVT_MENU, self.OnLoadAssembly, item1)
        itemR = menu.Append(-1, "Reload shape definition", "Reload shape definition")
        self.Bind(wx.EVT_MENU, self.OnReload, itemR)
        itemReset = menu.Append(-1, "Reset To shape definition", "Reset to shape definition")
        self.Bind(wx.EVT_MENU, self.OnReset, itemReset)
        menu.AppendSeparator()
        #itemR = menu.Append(-1, "&Reset to seed\tCtrl-R", "Reset system to the seed")
        #self.Bind(wx.EVT_MENU, self.OnReset, itemR)
        menu.AppendSeparator()
        itemSBT = menu.Append(-1, "Save shape definition", "Save shape definition")
        self.Bind(wx.EVT_MENU, self.OnSaveShapeDef, itemSBT)
        itemSBTA = menu.Append(-1, "Save shape definition as..", "Save shape definition as..")
        self.Bind(wx.EVT_MENU, self.OnSaveShapeDefAs, itemSBTA)
        menu.AppendSeparator()
        #itemSetSeed = menu.Append(-1, "Set current assembly as seed", "Set current assembly as seed assembly")
        #self.Bind(wx.EVT_MENU, self.OnSetSeed, itemSetSeed)
        #itemSS = menu.Append(-1, "Save seed", "Save seed assembly")
        #self.Bind(wx.EVT_MENU, self.OnSaveSeed, itemSS)
        #itemSSA = menu.Append(-1, "Save seed as..", "Save seed assembly as..")
        #self.Bind(wx.EVT_MENU, self.OnSaveSeedAs, itemSSA)
        #menu.AppendSeparator()
        #itemSA = menu.Append(-1, "Save current assembly", "Save current assembly")
        #self.Bind(wx.EVT_MENU, self.OnSaveAssembly, itemSA)
        #itemSAA = menu.Append(-1, "Save current assembly as..", "Save current assembly as..")
        #self.Bind(wx.EVT_MENU, self.OnSaveAssemblyAs, itemSAA)
        #menu.AppendSeparator()
        #item2 = menu.Append(-1, "&Save system definition and seed\tCtrl-S", "Save system definition and seed assembly")
        #self.Bind(wx.EVT_MENU, self.OnSaveAll, item2)
        menu.AppendSeparator()
        item3 = menu.Append(-1, "E&xit\tCtrl-X", "Exit program")
        self.Bind(wx.EVT_MENU, self.OnExitApp, item3)
        menuBar.Append(menu, "&File")

        menuCtrl = wx.Menu()
        #itemBP = menuCtrl.Append(-1, "&Breakpoints\tCtrl-B", "Set breakpoints")
        #self.Bind(wx.EVT_MENU, self.OnBreakpoints, itemBP)
        #menuCtrl.AppendSeparator()
        itemStep = menuCtrl.Append(-1, "&Forward\tCtrl-F", "Forward one step")
        self.Bind(wx.EVT_MENU, self.OnForward, itemStep)
        itemFF = menuCtrl.Append(-1, "&Go\tCtrl-G", "Fast forward")
        self.Bind(wx.EVT_MENU, self.OnGo, itemFF)
        #itemQuit = menuCtrl.Append(-1, "&Quit\tCtrl-Q", "Quit fast forward")
        #self.Bind(wx.EVT_MENU, self.OnQuit, itemQuit)
        menuBar.Append(menuCtrl, "&Control")

        self.menuView = wx.Menu()
        self.itemShowMess = self.menuView.Append(-1, "Show message window", "Toggle display of message window", kind=wx.ITEM_CHECK)
        self.Bind(wx.EVT_MENU, self.OnShowMessageWin, self.itemShowMess)
        self.menuView.Check(self.itemShowMess.GetId(), True)
        self.menuView.AppendSeparator()
        itemSTC = self.menuView.Append(-1, "&Center\tCtrl+C", "Scroll to center")
        self.Bind(wx.EVT_MENU, self.OnCenter, itemSTC)
        self.menuView.AppendSeparator()
        itemZIn = self.menuView.Append(-1, "Zoom in\tCtrl-+", "Zoom in")
        self.Bind(wx.EVT_MENU, self.OnZoomIn, itemZIn)
        itemZOut = self.menuView.Append(-1, "Zoom out\tCtrl--", "Zoom out")
        self.Bind(wx.EVT_MENU, self.OnZoomOut, itemZOut)
        self.menuView.AppendSeparator()
        itemTG = self.menuView.Append(-1, "&Toggle grid lines\tCtrl-T", "Toggle display of grid lines")
        self.Bind(wx.EVT_MENU, self.OnToggleGrid, itemTG)
        self.itemTM = self.menuView.Append(-1, "Update &messages\tCtrl-M", "Toggle display of messages", kind=wx.ITEM_CHECK)
        self.Bind(wx.EVT_MENU, self.OnMessageToggle, self.itemTM)
        self.menuView.Check(self.itemTM.GetId(), True)
        self.itemTD = self.menuView.Append(-1, "Update &display\tCtrl-D", "Toggle display of assembly updates", kind=wx.ITEM_CHECK)
        self.Bind(wx.EVT_MENU, self.OnDisplayToggle, self.itemTD)
        self.menuView.Check(self.itemTD.GetId(), True)
        menuBar.Append(self.menuView, "&View")

        #self.menuSettings = wx.Menu()
        #self.itemAdvColor = self.menuSettings.Append(-1, "Use advanced color picker", "Use advanced color picker dialog", kind=wx.ITEM_CHECK)
        #self.Bind(wx.EVT_MENU, self.OnAdvancedColorChecked, self.itemAdvColor)
        #self.menuSettings.Check(self.itemAdvColor.GetId(), False)
        #menuBar.Append(self.menuSettings, "&Settings")
    
        menu2 = wx.Menu()
        #itemDoc = menu2.Append(-1, "Documentation", "Show documentation")
        #self.Bind(wx.EVT_MENU, self.OnDoc, itemDoc)
        itemH = menu2.Append(-1, "&Help\tCtrl-H", "Show help")
        self.Bind(wx.EVT_MENU, self.OnHelp, itemH)
        menuBar.Append(menu2, "&Help")
        self.frame.SetMenuBar(menuBar)

        self.canvas = MyCanvas(self.frame, wx.ID_ANY, self.frame, wx.DefaultSize, self)
        self.frame._mgr.AddPane(self.canvas, wx.CENTER, 'Shape Definition Window')
        
        self.AddMessageWindow()
        
        self.frame.Bind(wx.EVT_CLOSE, self.OnCloseFrame)
        self.frame.Bind(wx.EVT_CHAR, self.canvas.OnChar)
        
        # Create toolbar
        tb1 = wx.ToolBar(self.frame, -1, wx.DefaultPosition, wx.DefaultSize, wx.TB_DOCKABLE | wx.TB_FLAT)
        tb1.SetToolSeparation(20)
        tb1.SetToolBitmapSize(wx.Size(24,24))
        tb1.AddLabelTool(ORIT_ID_NEW, "New System", wx.Image("images/new.xpm", wx.BITMAP_TYPE_XPM).ConvertToBitmap(), shortHelp="New System", longHelp="New oritatami system")
        tb1.AddLabelTool(ORIT_ID_LOAD, "Load System", wx.Image("images/load.xpm", wx.BITMAP_TYPE_XPM).ConvertToBitmap(), shortHelp="Load System", longHelp="Load oritatami system")
        tb1.AddLabelTool(ORIT_ID_SAVE, "Save System", wx.Image("images/save.xpm", wx.BITMAP_TYPE_XPM).ConvertToBitmap(), shortHelp="Save System", longHelp="Save oritatami system")
        tb1.AddLabelTool(ORIT_ID_RESET, "Reset To Shape", wx.Image("images/reset.xpm", wx.BITMAP_TYPE_XPM).ConvertToBitmap(), shortHelp="Reset To Shape Definition", longHelp="Reset to shape definition")
        tb1.Realize()
        tb1.Bind(wx.EVT_TOOL, self.OnTool)

        self.frame._mgr.AddPane(tb1, wx.aui.AuiPaneInfo().
                          Name("File toolbar").Caption("File Toolbar").
                          ToolbarPane().Top())
        #tb1.AddSeparator()
        #tb1.AddSeparator()
        
        tb1 = wx.ToolBar(self.frame, -1, wx.DefaultPosition, wx.DefaultSize, wx.TB_DOCKABLE | wx.TB_FLAT)
        tb1.SetToolSeparation(20)
        tb1.SetToolBitmapSize(wx.Size(24,24))
        btnImg = wx.Image("images/redo.xpm", wx.BITMAP_TYPE_XPM)
        btnImg.Rescale(25, 15)
        tb1.AddLabelTool(ORIT_ID_COMPILE, "Scale and rotate shape", btnImg.ConvertToBitmap(), shortHelp="Scale and rotate shape", longHelp="Scale and rotate shape")

        #tb1.AddLabelTool(ORIT_ID_FR, "Fast Rewind", wx.Image("images/fastleft.xpm", wx.BITMAP_TYPE_XPM).ConvertToBitmap(), shortHelp="Fast Rewind", longHelp="Fast rewind")
        #tb1.AddLabelTool(ORIT_ID_REV, "Rewind", wx.Image("images/left.xpm", wx.BITMAP_TYPE_XPM).ConvertToBitmap(), shortHelp="Rewind", longHelp="Rewind one step")
        #tb1.AddLabelTool(ORIT_ID_STOP, "Pause", wx.Image("images/pause.xpm", wx.BITMAP_TYPE_XPM).ConvertToBitmap(), shortHelp="Pause", longHelp="Pause simulation")
        tb1.AddLabelTool(ORIT_ID_FOR, "Forward (add one gadget)", wx.Image("images/right.xpm", wx.BITMAP_TYPE_XPM).ConvertToBitmap(), shortHelp="Forward (add one gadget)", longHelp="Forward one step")
        tb1.AddLabelTool(ORIT_ID_FF, "Fast Forward (add all gadgets)", wx.Image("images/fastright.xpm", wx.BITMAP_TYPE_XPM).ConvertToBitmap(), shortHelp="Fast Forward (add all gadgets)", longHelp="Fast forward")
        btnImg = wx.Image("images/logo.png", wx.BITMAP_TYPE_PNG)
        btnImg.Rescale(25, 15)
        tb1.AddLabelTool(ORIT_ID_GENOS, "Generate Oritatami system", btnImg.ConvertToBitmap(), shortHelp="Generate Oritatami system", longHelp="Generate Oritatami system")

        tb1.AddSeparator()
        tb1.AddLabelTool(ORIT_ID_ZOOMIN, "Zoom In", wx.Image("images/zoomin.xpm", wx.BITMAP_TYPE_XPM).ConvertToBitmap(), shortHelp="Zoom In", longHelp="Zoom in")
        tb1.AddLabelTool(ORIT_ID_ZOOMOUT, "Zoom Out", wx.Image("images/zoomout.xpm", wx.BITMAP_TYPE_XPM).ConvertToBitmap(), shortHelp="Zoom Out", longHelp="Zoom out")
        tb1.Realize()
        tb1.Bind(wx.EVT_TOOL, self.OnTool)

        self.frame._mgr.AddPane(tb1, wx.aui.AuiPaneInfo().
                          Name("Compilation toolbar").Caption("Compilation Toolbar").
                          ToolbarPane().Top())

        tb1 = wx.ToolBar(self.frame, -1, wx.DefaultPosition, wx.DefaultSize, wx.TB_DOCKABLE | wx.TB_FLAT)
        tb1.SetToolSeparation(10)
        #tb1.SetToolPacking(10)
        #tb1.SetToolBitmapSize(wx.Size(16,16))
        btnImg = wx.Image("images/add-remove-bead.png", wx.BITMAP_TYPE_PNG)
        btnImg.Rescale(25, 15)
        tb1.AddRadioTool(ORIT_ID_MODE_DRAW_BEAD, btnImg.ConvertToBitmap(), shortHelp="Add beads (left mouse)/remove beads (right mouse)", longHelp="Add beads (left mouse) and remove beads (right mouse)")

        btnImg = wx.Image("images/pan-zoom.png", wx.BITMAP_TYPE_PNG)
        btnImg.Rescale(45, 25)
        tb1.AddRadioTool(ORIT_ID_MODE_PAN_ZOOM, btnImg.ConvertToBitmap(), shortHelp="Pan (left mouse)/zoom (right mouse)", longHelp="Pan (left mouse) and zoom (right mouse)")
        btnImg.Rescale(25, 15)
        tb1.Realize()
        tb1.Bind(wx.EVT_TOOL, self.OnTool)
        
        self.SetMouseMode(ORIT_ID_MODE_DRAW_BEAD)

        self.frame._mgr.AddPane(tb1, wx.aui.AuiPaneInfo().
                          Name("Mouse mode toolbar").Caption("Mouse Mode Toolbar").
                          ToolbarPane().Top())
        

        # so set the frame to a good size for showing stuff
        self.frame.SetSize((1200, 800))
        self.frame.SetFocus()
        self.window = self.frame

        self.SetTopWindow(self.frame)
        #wx.Log_SetActiveTarget(wx.LogStderr())
        #wx.Log_SetTraceMask(wx.TraceMessages)
        
        # tell the manager to 'commit' all the changes just made
        self.frame._mgr.Update()
        #self.frame._mgr.OnFloatingPaneClosed()

        self.frame.Maximize()
        self.frame.Show(True)

        return True

    def AddMessageWindow(self):
        self.ctrlPanel = wx.Panel(self.frame, wx.ID_ANY, pos=wx.DefaultPosition, size=(300,100), style = wx.TAB_TRAVERSAL)
        self.messageWin = rt.RichTextCtrl(self.ctrlPanel, wx.ID_ANY, pos=wx.DefaultPosition, size=(500,70))#, style=wx.NOSCROLL|wx.HSCROLL|wx.NO_BORDER)
        self.messageWin.SetEditable(False)
        sz = wx.BoxSizer(wx.HORIZONTAL)
        sz.Add(self.messageWin,20,wx.EXPAND, 20)
        self.ctrlPanel.SetSizerAndFit(sz)
        self.frame._mgr.Bind(wx.aui.EVT_AUI_PANE_CLOSE, self.OnCloseMessagePane)
        self.frame._mgr.AddPane(self.ctrlPanel, wx.BOTTOM, 'Message Panel')
        self.frame._mgr.Update()

    def OnTool(self, evt):
        if evt.GetId() == ORIT_ID_MODE_PAN_ZOOM:
            self.SetMouseMode(ORIT_ID_MODE_PAN_ZOOM)
        elif evt.GetId() == ORIT_ID_MODE_DRAW_BEAD:
            self.SetMouseMode(ORIT_ID_MODE_DRAW_BEAD)
        elif evt.GetId() == ORIT_ID_NEW:
            self.canvas.DoNew()
        elif evt.GetId() == ORIT_ID_LOAD:
            self.canvas.DoLoad(bAssembly=True, bSystem=True)
        elif evt.GetId() == ORIT_ID_RESET:
            self.canvas.Reset()
        elif evt.GetId() == ORIT_ID_SAVE:
            self.canvas.DoSaveAll()
        elif evt.GetId() == ORIT_ID_COMPILE:
            self.canvas.Compile()
        elif evt.GetId() == ORIT_ID_GENOS:
            self.canvas.GenerateOritatamiSystem()
        elif evt.GetId() == ORIT_ID_FOR:
            self.canvas.DoForwardStep()
        elif evt.GetId() == ORIT_ID_FF:
            self.canvas.DoGo()
        elif evt.GetId() == ORIT_ID_ZOOMIN:
            self.canvas.ZoomThenRecenter(True)
        elif evt.GetId() == ORIT_ID_ZOOMOUT:
            self.canvas.ZoomThenRecenter(False)

    def OnCloseMessagePane(self, event):
        if event.GetPane().window == self.ctrlPanel:
            self.menuView.Check(self.itemShowMess.GetId(), False)
        event.Skip()
        
    def OnShowMessageWin(self, event):
        pane = self.frame._mgr.GetPane(self.ctrlPanel)
        if self.itemShowMess.IsChecked():
            pane.Show()
        else:
            pane.Hide()
        self.frame._mgr.Update()
        
    def OnAdvancedColorChecked(self, event):
        if self.itemShowMess.IsChecked():
            globalApp.useAdvancedColorPicker = True
        else:
            globalApp.useAdvancedColorPicker = False
        
    def SetMouseMode(self, mode):
        if not mode == self.canvas.mouseMode:
            self.canvas.linkStart = None
            self.canvas.linkEnd = None
            if (self.canvas.posSelected is not None) or (self.canvas.negSelected is not None):
                self.canvas.posSelected = None
                self.canvas.negSelected = None
                self.canvas.InitializeDrawing(self.canvas.pdc)
        if mode == ORIT_ID_MODE_PAN_ZOOM:
            self.canvas.mouseMode = ORIT_MODE_PAN_ZOOM
        elif mode == ORIT_ID_MODE_DRAW_BEAD:
            self.canvas.mouseMode = ORIT_MODE_DRAW_BEAD

    def SetOutputMessage(self, msg):
        if not self.showMessages:
            self.messageQueue.append(msg)
            return
        self.messageWin.SetInsertionPointEnd()
        self.messageWin.WriteText("\n" + msg)
        self.messageWin.SetInsertionPointEnd()
        self.messageWin.ScrollLines(self.messageWin.GetNumberOfLines()*5)
        #self.messageWin.Freeze()
        #self.messageWin.Clear()
        #self.messageWin.WriteText("\n" + msg)
        #self.messageWin.Thaw()
        #self.messageWin.ScrollLines(-1)

    def ShowErrorDialog(self, msg):
        dlg = wx.MessageDialog(self.frame, msg, caption="Error", style=wx.ICON_ERROR|wx.OK)
        dlg.ShowModal()
        dlg.Destroy()
        
    def ShowWarningDialog(self, msg):
        dlg = wx.MessageDialog(self.frame, msg, caption="Warning", style=wx.ICON_WARNING|wx.OK)
        dlg.ShowModal()
        dlg.Destroy()

    def OnExitApp(self, evt):
        self.frame.Close(True)

    def OnNew(self, evt):
        self.canvas.DoNew()

    def OnLoadAssembly(self, evt):
        self.canvas.DoLoad(bAssembly=True, bSystem=False)
        
    def OnLoadSystem(self, evt):
        self.canvas.DoLoad(bAssembly=False, bSystem=True)

    def OnReload(self, evt):
        self.canvas.Reload()
        
    def OnReset(self, evt):
        self.canvas.Reset()
        
    def OnCompile(self, evt):
        self.canvas.Compile()

    def OnGenerateOritatamiSystem(self, evt):
        self.canvas.GenerateOritatamiSystem()

    def OnSaveShapeDef(self, event):
        self.canvas.DoSaveShapeDef()

    def OnSaveShapeDefAs(self, event):
        self.canvas.DoSaveShapeDef(bSaveAs=True)

    def OnSetSeed(self, event):
        self.canvas.DoSetSeed()

    def OnSaveSeed(self, event):
        self.canvas.DoSaveAssembly(bOnlySeed=True)

    def OnSaveSeedAs(self, event):
        self.canvas.DoSaveAssembly(bOnlySeed=True, bSaveAs=True)

    def OnSaveAssembly(self, event):
        self.canvas.DoSaveAssembly()

    def OnSaveAssemblyAs(self, event):
        self.canvas.DoSaveAssembly(bOnlySeed=False, bSaveAs=True)

    def OnSaveAll(self, event):
        self.canvas.DoSaveAll()
        
    def OnForward(self, event):
        self.canvas.DoForwardStep()
        
    def OnGo(self, event):
        self.canvas.DoGo()
        
    def OnCenter(self, event):
        self.canvas.ScrollToCoords((0,0))
        
    def OnZoomIn(self, event):
        self.canvas.ZoomThenRecenter(True)

    def OnZoomOut(self, event):
        self.canvas.ZoomThenRecenter(False)
        
    def OnToggleGrid(self, event):
        self.canvas.ToggleGrid()
    
    def OnMessageToggle(self, event):
        if self.itemTM.IsChecked():
            self.canvas.MessageToggle(True)
        else:
            self.canvas.MessageToggle(False)
        
    def OnDisplayToggle(self, event):
        if self.itemTD.IsChecked():
            self.canvas.DisplayToggle(True)
        else:
            self.canvas.DisplayToggle(False)

    def OnDoc(self, event):
        if self.docFrame is None:
            self.docFrame = DocumentationFrame(self)
        self.docFrame.SetFocus()

    def OnHelp(self, event):
        #info = wx.AboutDialogInfo()
        #info.SetName("Oritatami Simulator")
        #info.SetVersion("0.1.0")
        helpStr           = "                                         Oritatami Shape Maker\n"
        helpStr = helpStr + "                                           version 0.1.0\n\n"
        helpStr = helpStr + " This program allows a user to load or draw a shape, and then generate\n"
        helpStr = helpStr + " an Oritatami system which will self-assemble a scaled version of it in\n"
        helpStr = helpStr + " the OritatamiSim simulator.\n\n"
        helpStr = helpStr + " To generate a system, first use the mouse to select the connected set of\n"
        helpStr = helpStr + " points defining the shape.  Then, hit the 'Scale and rotate shape' button.\n"
        helpStr = helpStr + " If the points are connected, this will succeed in creating a scaled and\n"
        helpStr = helpStr + " slightly rotated version of the points.  Then, hit the 'Forward' or\n"
        helpStr = helpStr + " 'Fast Forward' button until all points in the scaled and rotated shape\n"
        helpStr = helpStr + " have been replaced by hexagons of seven points, and a Hamiltonian cycle\n"
        helpStr = helpStr + " has been drawn through all included points.  Finally, hit the 'Generate\n"
        helpStr = helpStr + " Oritatami system' button to select the names of the output files.\n\n"
        helpStr = helpStr + "    List of commands:\n\n"
        helpStr = helpStr + "      l : ('load') select shape definition file to load\n"
        helpStr = helpStr + "      s : save current shape definition\n"
        helpStr = helpStr + "      r : ('reset') reset current system back to shape definition\n"
        helpStr = helpStr + "      f : ('forward') simulate one step\n"
        helpStr = helpStr + "      g : ('go') begin fast forward simulation\n"
        helpStr = helpStr + "      q : ('quit') quit fast forward simulation\n"
        helpStr = helpStr + "      m : ('messges') toggle display of output messages\n"
        helpStr = helpStr + "      d : ('display') toggle updates to the display of the assembly\n"
        helpStr = helpStr + "      t : ('toggle') toggle display of grid lines\n"
        helpStr = helpStr + "      c : re-center grid\n"
        #helpStr = helpStr + "      b : ('breakpoints') set breakpoints\n"
        helpStr = helpStr + "      +/-: zoom in/out\n"
        helpStr = helpStr + "      <left mouse click>:  add point to shape definition\n"
        helpStr = helpStr + "      <right mouse click>:  remove point from shape definition\n"
        helpStr = helpStr + "      h : ('help') show this help dialog"
        helpStr = helpStr + "\n\n"
        #info.SetDescription(helpStr)
        #wx.AboutBox(info)
        helpPanel = wx.Panel(self.frame)
        helpText = wx.StaticText(helpPanel, label=helpStr)
        helpText.SetBackgroundColour((233,233,233))
        # add the panes to the manager
        self.frame._mgr.AddPane(helpText, wx.aui.AuiPaneInfo().Float().Caption("Oritatami Shape Maker Help"))
        self.frame._mgr.Update()
        

    def OnCloseFrame(self, evt):

        if not self.docFrame is None:
            wx.PostEvent(self.docFrame, evt)

        # deinitialize the frame manager
        self.frame._mgr.UnInit()
        # delete the frame
        self.frame.Destroy()

#----------------------------------------------------------------------------
        
    def OnDlgButton(self, event):
        if event.LeftDown():
            if event.EventObject == self.resetBtn:
                self.EndModal(ORIT_ID_RESET_BEADSEQ)
            if event.EventObject == self.reloadBtn:
                self.EndModal(ORIT_ID_RELOAD_BEADSEQ)
            if event.EventObject == self.cancelBtn:
                self.EndModal(ORIT_ID_CANCEL_BEADSEQ)
        
#----------------------------------------------------------------------------
class DontShowAgainDialog(wx.Dialog):
    def __init__(
            self, parent, ID, title, message, size=wx.DefaultSize, pos=wx.DefaultPosition, style=wx.DEFAULT_DIALOG_STYLE):

        self.bShowAgain = True
        # Instead of calling wx.Dialog.__init__ we precreate the dialog
        # so we can set an extra style that must be set before
        # creation, and then we create the GUI object using the Create
        # method.
        pre = wx.PreDialog()
        pre.Create(parent, ID, title, pos, size, style)

        # This next step is the most important, it turns this Python
        # object into the real wrapper of the dialog (instead of pre)
        # as far as the wxPython extension is concerned.
        self.PostCreate(pre)

        # Now continue with the normal construction of the dialog
        # contents
        sizer = wx.BoxSizer(wx.VERTICAL)

        label = wx.StaticText(self, -1, message)
        sizer.Add(label, 0, wx.ALIGN_CENTRE|wx.ALL, 5)

        self.checkbox = wx.CheckBox(self, -1, "Do not show this warning again")
        sizer.Add(self.checkbox, 0, wx.ALIGN_RIGHT|wx.ALL, 5)

        btnsizer = wx.StdDialogButtonSizer()
        
        self.okBtn = wx.Button(self, wx.ID_OK, "OK")
        self.okBtn.Bind(wx.EVT_MOUSE_EVENTS, self.OnDlgButton)
        self.okBtn.SetDefault()
        btnsizer.AddButton(self.okBtn)
        
        self.cancelBtn = wx.Button(self, wx.ID_CANCEL, "Cancel")
        self.cancelBtn.Bind(wx.EVT_MOUSE_EVENTS, self.OnDlgButton)
        btnsizer.AddButton(self.cancelBtn)

        btnsizer.Realize()

        sizer.Add(btnsizer, 0, wx.ALIGN_CENTER_VERTICAL|wx.ALIGN_RIGHT|wx.ALL, 5)

        self.SetSizer(sizer)
        sizer.Fit(self)
        
    def OnDlgButton(self, event):
        if event.LeftDown():
            bChecked = self.checkbox.GetValue()
            if event.EventObject == self.okBtn:
                if bChecked == True:
                    self.EndModal(ORIT_ID_OK_CHECK)
                else:
                    self.EndModal(ORIT_ID_OK_NOCHECK)
            if event.EventObject == self.cancelBtn:
                self.EndModal(ORIT_ID_CANCEL)
        
#----------------------------------------------------------------------------

class SomeCustomDialog(wx.Dialog):
    def __init__(
            self, parent, ID, size=wx.DefaultSize, pos=wx.DefaultPosition, 
            style=wx.DEFAULT_DIALOG_STYLE,
            useMetal=False,
            ):

        title = "Custom dialog"
        # Instead of calling wx.Dialog.__init__ we precreate the dialog
        # so we can set an extra style that must be set before
        # creation, and then we create the GUI object using the Create
        # method.
        pre = wx.PreDialog()
        pre.SetExtraStyle(wx.DIALOG_EX_CONTEXTHELP)
        pre.Create(parent, ID, title, pos, size, style)

        # This next step is the most important, it turns this Python
        # object into the real wrapper of the dialog (instead of pre)
        # as far as the wxPython extension is concerned.
        self.PostCreate(pre)

        # This extra style can be set after the UI object has been created.
        if 'wxMac' in wx.PlatformInfo and useMetal:
            self.SetExtraStyle(wx.DIALOG_EX_METAL)


        # Now continue with the normal construction of the dialog
        # contents
        sizer = wx.BoxSizer(wx.VERTICAL)

        label = wx.StaticText(self, -1, "This is a wx.Dialog")
        label.SetHelpText("This is the help text for the label")
        sizer.Add(label, 0, wx.ALIGN_CENTRE|wx.ALL, 5)

        box = wx.BoxSizer(wx.HORIZONTAL)

        label = wx.StaticText(self, -1, "Field #1:")
        label.SetHelpText("This is the help text for the label")
        box.Add(label, 0, wx.ALIGN_CENTRE|wx.ALL, 5)

        text = wx.TextCtrl(self, -1, "", size=(80,-1))
        text.SetHelpText("Here's some help text for field #1")
        box.Add(text, 1, wx.ALIGN_CENTRE|wx.ALL, 5)

        sizer.Add(box, 0, wx.GROW|wx.ALIGN_CENTER_VERTICAL|wx.ALL, 5)

        box = wx.BoxSizer(wx.HORIZONTAL)

        label = wx.StaticText(self, -1, "Field #2:")
        label.SetHelpText("This is the help text for the label")
        box.Add(label, 0, wx.ALIGN_CENTRE|wx.ALL, 5)

        text = wx.TextCtrl(self, -1, "", size=(80,-1))
        text.SetHelpText("Here's some help text for field #2")
        box.Add(text, 1, wx.ALIGN_CENTRE|wx.ALL, 5)

        sizer.Add(box, 0, wx.GROW|wx.ALIGN_CENTER_VERTICAL|wx.ALL, 5)

        line = wx.StaticLine(self, -1, size=(20,-1), style=wx.LI_HORIZONTAL)
        sizer.Add(line, 0, wx.GROW|wx.ALIGN_CENTER_VERTICAL|wx.RIGHT|wx.TOP, 5)

        btnsizer = wx.StdDialogButtonSizer()
        
        if wx.Platform != "__WXMSW__":
            btn = wx.ContextHelpButton(self)
            btnsizer.AddButton(btn)
        
        btn = wx.Button(self, wx.ID_OK)
        btn.SetHelpText("The OK button completes the dialog")
        btn.SetDefault()
        btnsizer.AddButton(btn)

        btn = wx.Button(self, wx.ID_CANCEL)
        btn.SetHelpText("The Cancel button cancels the dialog. (Cool, huh?)")
        btnsizer.AddButton(btn)
        btnsizer.Realize()

        sizer.Add(btnsizer, 0, wx.ALIGN_CENTER_VERTICAL|wx.ALL, 5)

        self.SetSizer(sizer)
        sizer.Fit(self)

#----------------------------------------------------------------------------

class DocumentationTB(wx.Treebook):
    def __init__(self, parent, idIn):
        wx.Treebook.__init__(self, parent, idIn, style=
                             wx.BK_DEFAULT
                             #wx.BK_TOP
                             #wx.BK_BOTTOM
                             #wx.BK_LEFT
                             #wx.BK_RIGHT
                            )
        
        self.windows = []
        # Now make a bunch of panels for the list book
        for fileName in DocumentationFileList:
            win = self.MakeDocPanel(fileName)
            self.AddPage(win, fileName)
            self.windows.append(win)
            #win = self.MakeDocPanel(colour)
            #self.AddSubPage(win, 'a sub-page')

        # This is a workaround for a sizing bug on Mac...
        #wx.FutureCall(100, self.AdjustSize)

        self.Bind(wx.EVT_TREEBOOK_PAGE_CHANGING, self.OnPageChanging)

    def AdjustSize(self):
        #print self.GetTreeCtrl().GetBestSize()
        self.GetTreeCtrl().InvalidateBestSize()
        self.SendSizeEvent()
        #print self.GetTreeCtrl().GetBestSize()
        

    def MakeDocPanel(self, fileName):
        p = wx.html.HtmlWindow(self) 
        if "gtk2" in wx.PlatformInfo: 
            p.SetStandardFonts() 
        p.LoadPage("docs/" + fileName + ".html")
        return p

    def OnPageChanging(self, event):
        selected = event.GetSelection()
        #print("new: " + str(selected))
        correctDoc = DocumentationFileList[selected]
        self.windows[selected].LoadPage("docs/" + correctDoc + ".html")
        event.Skip()

#----------------------------------------------------------------------------

class DocumentationFrame(wx.Frame):
    def __init__(self, app):
        wx.Frame.__init__(self, None, -1, "Oritatami Simulator Documentation", pos=(50,50), size=(400,400),
                         style=wx.DEFAULT_FRAME_STYLE, name="Oritatami Simulator Documentation")
        self.app = app
        self._mgr = wx.aui.AuiManager(self)
        self.SetIcon(wx.IconFromBitmap(wx.Bitmap("images/logo.png", wx.BITMAP_TYPE_PNG)))

        self.treewin = DocumentationTB(self, -1)
        self._mgr.AddPane(self.treewin, wx.CENTER, 'Documentation Window')
        
        self.Bind(wx.EVT_CLOSE, self.OnCloseDocFrame)
        self.CreateStatusBar(3)
        self.SetSize((1200, 800))
        self.SetFocus()

        self._mgr.Update()
        #self.Maximize()
        self.Show(True)
        
    def OnCloseDocFrame(self, evt):
        # deinitialize the frame manager
        self._mgr.UnInit()
        self.app.docFrame = None
        # delete the frame
        self.Destroy()

#----------------------------------------------------------------------------
#  Create the application and run it

app = OritatamiSimulatorApp()
app.MainLoop()

#---------------------------------------------------------------------------

